"""workflow_templates table updates oss

Revision ID: b4738bd17198
Revises: 7ab8e817802a
Create Date: 2025-12-12 02:32:15.078365+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "b4738bd17198"
down_revision: Union[str, None] = "7ab8e817802a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "workflow_templates",
        sa.Column("workflow_template_id", sa.String(), nullable=False),
        sa.Column("workflow_permanent_id", sa.String(), nullable=False),
        sa.Column("organization_id", sa.String(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("modified_at", sa.DateTime(), nullable=False),
        sa.Column("deleted_at", sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(
            ["organization_id"],
            ["organizations.organization_id"],
        ),
        sa.PrimaryKeyConstraint("workflow_template_id"),
    )
    op.create_index(
        op.f("ix_workflow_templates_organization_id"), "workflow_templates", ["organization_id"], unique=False
    )
    op.create_index(
        op.f("ix_workflow_templates_workflow_permanent_id"),
        "workflow_templates",
        ["workflow_permanent_id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_workflow_templates_workflow_permanent_id"), table_name="workflow_templates")
    op.drop_index(op.f("ix_workflow_templates_organization_id"), table_name="workflow_templates")
    op.drop_table("workflow_templates")
    # ### end Alembic commands ###



from time import perf_counter

import httpx
import structlog
from fastapi import Depends, HTTPException, status

from skyvern.exceptions import (
    BlockedHost,
    MissingApiKey,
    MissingWebhookTarget,
    SkyvernHTTPException,
    TaskNotFound,
    WebhookReplayError,
    WorkflowRunNotFound,
)
from skyvern.forge import app
from skyvern.forge.sdk.core.security import generate_skyvern_webhook_signature
from skyvern.forge.sdk.db.enums import OrganizationAuthTokenType
from skyvern.forge.sdk.routes.routers import base_router, legacy_base_router
from skyvern.forge.sdk.schemas.organizations import Organization
from skyvern.forge.sdk.services import org_auth_service
from skyvern.schemas.webhooks import (
    RunWebhookPreviewResponse,
    RunWebhookReplayRequest,
    RunWebhookReplayResponse,
    TestWebhookRequest,
    TestWebhookResponse,
)
from skyvern.services.webhook_service import (
    build_run_preview,
    build_sample_task_payload,
    build_sample_workflow_run_payload,
    replay_run_webhook,
)
from skyvern.utils.url_validators import validate_url

LOG = structlog.get_logger()


@legacy_base_router.post(
    "/internal/test-webhook",
    tags=["Internal"],
    description="Test a webhook endpoint by sending a sample payload",
    summary="Test webhook endpoint",
    include_in_schema=False,
)
@base_router.post(
    "/internal/test-webhook",
    tags=["Internal"],
    description="Test a webhook endpoint by sending a sample payload",
    summary="Test webhook endpoint",
    include_in_schema=False,
)
async def test_webhook(
    request: TestWebhookRequest,
    current_org: Organization = Depends(org_auth_service.get_current_org),
) -> TestWebhookResponse:
    """
    Test a webhook endpoint by sending a sample signed payload.

    This endpoint allows users to:
    - Validate their webhook receiver can be reached
    - Test HMAC signature verification
    - See the exact headers and payload format Skyvern sends

    The endpoint respects SSRF protection (BLOCKED_HOSTS, private IPs) and will return
    a helpful error message if the URL is blocked.
    """
    start_time = perf_counter()

    # Validate the URL (raises BlockedHost or SkyvernHTTPException for invalid URLs)
    try:
        validated_url = validate_url(request.webhook_url)
        if not validated_url:
            return TestWebhookResponse(
                status_code=None,
                latency_ms=0,
                response_body="",
                headers_sent={},
                error="Invalid webhook URL",
            )
    except BlockedHost as exc:
        blocked_host: str | None = getattr(exc, "host", None)
        return TestWebhookResponse(
            status_code=None,
            latency_ms=0,
            response_body="",
            headers_sent={},
            error=(
                f"This URL is blocked by SSRF protection (host: {blocked_host or 'unknown'}). "
                "Add the host to ALLOWED_HOSTS to test internal endpoints or use an external receiver "
                "such as webhook.site or requestbin.com."
            ),
        )
    except SkyvernHTTPException as exc:
        error_message = getattr(exc, "message", None) or "Invalid webhook URL. Use http(s) and a valid host."
        return TestWebhookResponse(
            status_code=None,
            latency_ms=0,
            response_body="",
            headers_sent={},
            error=error_message,
        )
    except Exception as exc:  # pragma: no cover - defensive guard
        LOG.exception("Error validating webhook URL", error=str(exc), webhook_url=request.webhook_url)
        return TestWebhookResponse(
            status_code=None,
            latency_ms=0,
            response_body="",
            headers_sent={},
            error="Unexpected error while validating the webhook URL.",
        )

    # Build the sample payload based on run type
    try:
        if request.run_type == "task":
            payload = build_sample_task_payload(run_id=request.run_id)
        else:  # workflow_run
            payload = build_sample_workflow_run_payload(run_id=request.run_id)
    except Exception as e:
        LOG.exception("Error building sample payload", error=str(e), run_type=request.run_type)
        return TestWebhookResponse(
            status_code=None,
            latency_ms=0,
            response_body="",
            headers_sent={},
            error=f"Failed to build sample payload: {str(e)}",
        )

    # Get the organization's API key to sign the webhook
    # For testing, we use a placeholder if no API key is available
    api_key_obj = await app.DATABASE.get_valid_org_auth_token(
        current_org.organization_id,
        OrganizationAuthTokenType.api.value,
    )
    api_key = api_key_obj.token if api_key_obj else "test_api_key_placeholder"

    signed_data = generate_skyvern_webhook_signature(payload=payload, api_key=api_key)

    # Send the webhook request
    status_code = None
    response_body = ""
    error = None

    try:
        async with httpx.AsyncClient() as client:
            response = await client.post(
                validated_url,
                content=signed_data.signed_payload,
                headers=signed_data.headers,
                timeout=httpx.Timeout(60.0),
            )
            status_code = response.status_code

            # Capture first 2KB of response body
            response_text = response.text
            if len(response_text) > 2048:
                response_body = response_text[:2048] + "\n... (truncated)"
            else:
                response_body = response_text

    except httpx.TimeoutException:
        error = "Request timed out after 60 seconds."
        LOG.warning(
            "Test webhook timeout",
            organization_id=current_org.organization_id,
            webhook_url=validated_url,
        )
    except httpx.NetworkError as exc:
        error = f"Could not reach URL: {exc}"
        LOG.warning(
            "Test webhook network error",
            organization_id=current_org.organization_id,
            webhook_url=validated_url,
            error=str(exc),
        )
    except Exception as exc:  # pragma: no cover - defensive guard
        error = f"Unexpected error: {exc}"
        LOG.error(
            "Test webhook unexpected error",
            organization_id=current_org.organization_id,
            webhook_url=validated_url,
            error=str(exc),
            exc_info=True,
        )

    latency_ms = int((perf_counter() - start_time) * 1000)

    return TestWebhookResponse(
        status_code=status_code,
        latency_ms=latency_ms,
        response_body=response_body,
        headers_sent=signed_data.headers,
        error=error,
    )


@legacy_base_router.get(
    "/internal/runs/{run_id}/test-webhook",
    tags=["Internal"],
    response_model=RunWebhookPreviewResponse,
    summary="Preview webhook replay payload",
    include_in_schema=False,
)
@base_router.get(
    "/internal/runs/{run_id}/test-webhook",
    tags=["Internal"],
    response_model=RunWebhookPreviewResponse,
    summary="Preview webhook replay payload",
    include_in_schema=False,
)
async def preview_webhook_replay(
    run_id: str,
    current_org: Organization = Depends(org_auth_service.get_current_org),  # noqa: B008
) -> RunWebhookPreviewResponse:
    """Return the replay payload preview for a completed run.

    Args:
        run_id (str): Identifier of the run to preview.
        current_org (Organization): Organization context for permission and signing.

    Returns:
        RunWebhookPreviewResponse: Payload and headers that would be used to replay the webhook.

    Raises:
        HTTPException: 400 if the organization lacks a valid API key or other replay preconditions fail.
        HTTPException: 404 if the specified run cannot be found.
        HTTPException: 500 if an unexpected error occurs while building the preview.
    """
    try:
        return await build_run_preview(
            organization_id=current_org.organization_id,
            run_id=run_id,
        )
    except MissingApiKey as exc:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(exc)) from exc
    except (TaskNotFound, WorkflowRunNotFound) as exc:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(exc)) from exc
    except WebhookReplayError as exc:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(exc)) from exc
    except SkyvernHTTPException as exc:
        raise HTTPException(status_code=exc.status_code, detail=str(exc)) from exc
    except Exception as exc:  # pragma: no cover - defensive guard
        LOG.error("Failed to build webhook replay preview", run_id=run_id, error=str(exc), exc_info=True)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Failed to build webhook replay preview.",
        ) from exc


@legacy_base_router.post(
    "/internal/runs/{run_id}/test-webhook",
    tags=["Internal"],
    response_model=RunWebhookReplayResponse,
    summary="Replay webhook for a completed run",
    include_in_schema=False,
)
@base_router.post(
    "/internal/runs/{run_id}/test-webhook",
    tags=["Internal"],
    response_model=RunWebhookReplayResponse,
    summary="Replay webhook for a completed run",
    include_in_schema=False,
)
async def trigger_webhook_replay(
    run_id: str,
    request: RunWebhookReplayRequest,
    current_org: Organization = Depends(org_auth_service.get_current_org),  # noqa: B008
) -> RunWebhookReplayResponse:
    """Replay a completed run's webhook to the stored or override URL.

    Args:
        run_id (str): Identifier of the run whose webhook should be replayed.
        request (RunWebhookReplayRequest): Optional override URL details for the replay.
        current_org (Organization): Organization context for permission and signing.

    Returns:
        RunWebhookReplayResponse: Delivery status information for the replay attempt.

    Raises:
        HTTPException: 400 if no target URL is available, the organization lacks an API key, or replay validation fails.
        HTTPException: 404 if the specified run cannot be found.
        HTTPException: 500 if an unexpected error occurs while replaying the webhook.
    """
    try:
        return await replay_run_webhook(
            organization_id=current_org.organization_id,
            run_id=run_id,
            target_url=request.override_webhook_url,
        )
    except MissingWebhookTarget as exc:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(exc)) from exc
    except MissingApiKey as exc:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(exc)) from exc
    except (TaskNotFound, WorkflowRunNotFound) as exc:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=str(exc)) from exc
    except WebhookReplayError as exc:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(exc)) from exc
    except SkyvernHTTPException as exc:
        raise HTTPException(status_code=exc.status_code, detail=str(exc)) from exc
    except Exception as exc:  # pragma: no cover - defensive guard
        LOG.error("Failed to replay webhook", run_id=run_id, error=str(exc), exc_info=True)
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="Failed to replay webhook.",
        ) from exc



"""
Provides WS endpoints for streaming messages to/from our frontend application.
"""

import structlog
from fastapi import WebSocket

from skyvern.config import settings
from skyvern.forge.sdk.routes.routers import base_router, legacy_base_router
from skyvern.forge.sdk.routes.streaming.auth import _auth as local_auth
from skyvern.forge.sdk.routes.streaming.auth import auth as real_auth
from skyvern.forge.sdk.routes.streaming.channels.message import (
    Loops,
    MessageChannel,
    get_message_channel_for_browser_session,
    get_message_channel_for_workflow_run,
)
from skyvern.forge.sdk.utils.aio import collect

LOG = structlog.get_logger()


@base_router.websocket("/stream/messages/browser_session/{browser_session_id}")
async def browser_session_messages(
    websocket: WebSocket,
    browser_session_id: str,
    apikey: str | None = None,
    client_id: str | None = None,
    token: str | None = None,
) -> None:
    return await messages(
        websocket=websocket,
        browser_session_id=browser_session_id,
        apikey=apikey,
        client_id=client_id,
        token=token,
    )


@legacy_base_router.websocket("/stream/messages/workflow_run/{workflow_run_id}")
async def workflow_run_messages(
    websocket: WebSocket,
    workflow_run_id: str,
    apikey: str | None = None,
    client_id: str | None = None,
    token: str | None = None,
) -> None:
    return await messages(
        websocket=websocket,
        workflow_run_id=workflow_run_id,
        apikey=apikey,
        client_id=client_id,
        token=token,
    )


async def messages(
    websocket: WebSocket,
    browser_session_id: str | None = None,
    workflow_run_id: str | None = None,
    apikey: str | None = None,
    client_id: str | None = None,
    token: str | None = None,
) -> None:
    auth = local_auth if settings.ENV == "local" else real_auth
    organization_id = await auth(apikey=apikey, token=token, websocket=websocket)

    if not organization_id:
        LOG.warning(
            "Authentication failed.",
            browser_session_id=browser_session_id,
            workflow_run_id=workflow_run_id,
        )
        return

    if not client_id:
        LOG.error(
            "No client ID provided.",
            browser_session_id=browser_session_id,
            workflow_run_id=workflow_run_id,
        )
        await websocket.close(code=1002)
        return

    message_channel: MessageChannel
    loops: Loops = []

    if browser_session_id:
        result = await get_message_channel_for_browser_session(
            client_id=client_id,
            browser_session_id=browser_session_id,
            organization_id=organization_id,
            websocket=websocket,
        )
    elif workflow_run_id:
        result = await get_message_channel_for_workflow_run(
            client_id=client_id,
            workflow_run_id=workflow_run_id,
            organization_id=organization_id,
            websocket=websocket,
        )
    else:
        LOG.error(
            "[WS] messages: no browser_session_id or workflow_run_id provided",
            client_id=client_id,
            organization_id=organization_id,
        )
        await websocket.close(code=1002)
        return

    if not result:
        LOG.debug(
            "No message channel found.",
            browser_session_id=browser_session_id,
            workflow_run_id=workflow_run_id,
            organization_id=organization_id,
        )
        await websocket.close(code=1013)
        return

    message_channel, loops = result

    try:
        LOG.info(
            "Starting message channel loops.",
            browser_session_id=browser_session_id,
            workflow_run_id=workflow_run_id,
            organization_id=organization_id,
        )
        await collect(loops)
    except Exception:
        LOG.exception(
            "An exception occurred in the message loop function(s).",
            browser_session_id=browser_session_id,
            workflow_run_id=workflow_run_id,
            organization_id=organization_id,
        )
    finally:
        LOG.info(
            "Closing the message channel.",
            browser_session_id=browser_session_id,
            workflow_run_id=workflow_run_id,
            organization_id=organization_id,
        )
        await message_channel.close(reason="message-stream-closed")



from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable, ParamSpec, TypeVar

P = ParamSpec("P")
R = TypeVar("R")


class BaseTrace(ABC):
    @abstractmethod
    def traced(
        self,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
        tags: list[str] | None = None,
        **kwargs: Any,
    ) -> Callable[[Callable[P, R]], Callable[P, R]]:
        pass

    @abstractmethod
    def traced_async(
        self,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
        tags: list[str] | None = None,
        **kwargs: Any,
    ) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
        pass

    def add_task_completion_tag(self, status: str) -> None:
        """Add a completion tag to the current trace based on task/workflow status."""

    def add_experiment_metadata(self, experiment_data: dict[str, Any]) -> None:
        """Add experiment metadata to the current trace."""


class NoOpTrace(BaseTrace):
    def traced(
        self,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
        tags: list[str] | None = None,
        **kwargs: Any,
    ) -> Callable[[Callable[P, R]], Callable[P, R]]:
        return lambda func: func

    def traced_async(
        self,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
        tags: list[str] | None = None,
        **kwargs: Any,
    ) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
        return lambda func: func



import asyncio

from structlog import get_logger

from skyvern.forge.sdk.db.agent_db import AgentDB
from skyvern.forge.sdk.schemas.persistent_browser_sessions import PersistentBrowserSession

LOG = get_logger(__name__)


async def wait_on_persistent_browser_address(
    db: AgentDB,
    session_id: str,
    organization_id: str,
    timeout: int = 600,
    poll_interval: float = 2,
) -> str | None:
    persistent_browser_session = await await_browser_session(db, session_id, organization_id, timeout, poll_interval)
    return persistent_browser_session.browser_address if persistent_browser_session else None


async def await_browser_session(
    db: AgentDB,
    session_id: str,
    organization_id: str,
    timeout: int = 600,
    poll_interval: float = 2,
) -> PersistentBrowserSession | None:
    try:
        async with asyncio.timeout(timeout):
            while True:
                persistent_browser_session = await db.get_persistent_browser_session(session_id, organization_id)
                if persistent_browser_session is None:
                    raise Exception(f"Persistent browser session not found for {session_id}")

                LOG.debug(
                    "Checking browser address",
                    session_id=session_id,
                    address=persistent_browser_session.browser_address,
                )

                if persistent_browser_session.browser_address:
                    return persistent_browser_session

                await asyncio.sleep(poll_interval)
    except asyncio.TimeoutError:
        LOG.warning(f"Browser address not found for persistent browser session {session_id}")

    return None



from contextvars import ContextVar
from dataclasses import dataclass, field
from typing import Any
from zoneinfo import ZoneInfo

from playwright.async_api import Frame, Page


@dataclass
class SkyvernContext:
    request_id: str | None = None
    organization_id: str | None = None
    organization_name: str | None = None
    task_id: str | None = None
    step_id: str | None = None
    workflow_id: str | None = None
    workflow_permanent_id: str | None = None
    workflow_run_id: str | None = None
    root_workflow_run_id: str | None = None
    task_v2_id: str | None = None
    max_steps_override: int | None = None
    browser_session_id: str | None = None
    tz_info: ZoneInfo | None = None
    run_id: str | None = None
    totp_codes: dict[str, str | None] = field(default_factory=dict)
    log: list[dict] = field(default_factory=list)
    hashed_href_map: dict[str, str] = field(default_factory=dict)
    refresh_working_page: bool = False
    frame_index_map: dict[Frame, int] = field(default_factory=dict)
    dropped_css_svg_element_map: dict[str, bool] = field(default_factory=dict)
    max_screenshot_scrolls: int | None = None
    browser_container_ip: str | None = None
    browser_container_task_arn: str | None = None

    # feature flags
    enable_parse_select_in_extract: bool = False
    use_prompt_caching: bool = False
    cached_static_prompt: str | None = None
    vertex_cache_name: str | None = None  # Vertex AI cache resource name for explicit caching
    vertex_cache_key: str | None = None  # Logical cache key (includes variant + llm key)
    vertex_cache_variant: str | None = None  # Variant identifier used when creating the cache
    prompt_caching_settings: dict[str, bool] | None = None
    enable_speed_optimizations: bool = False

    # script run context
    script_id: str | None = None
    script_revision_id: str | None = None
    action_order: int = 0
    prompt: str | None = None
    parent_workflow_run_block_id: str | None = None
    loop_metadata: dict[str, Any] | None = None
    loop_output_values: list[dict[str, Any]] | None = None
    script_run_parameters: dict[str, Any] = field(default_factory=dict)
    script_mode: bool = False
    ai_mode_override: str | None = None

    # magic link handling
    # task_id is the key, page is the value
    # we only consider the page is a magic link page in the same task scope
    # for example, login block has a magic link page,
    # but it will only be considered as a magic link page in the login block scope
    # next blocks won't consider the page as a magic link page
    magic_link_pages: dict[str, Page] = field(default_factory=dict)

    # parallel verification optimization
    # stores pre-scraped data for next step to avoid re-scraping
    next_step_pre_scraped_data: dict[str, Any] | None = None
    speculative_plans: dict[str, Any] = field(default_factory=dict)

    """
    Example output value:
    {"loop_value": "str", "output_parameter": "the key of the parameter", "output_value": Any}
    """
    generate_script: bool = True
    action_ai_overrides: dict[str, dict[int, str]] = field(default_factory=dict)
    action_counters: dict[str, int] = field(default_factory=dict)

    # Track if script generation skipped any actions due to missing data (race condition)
    # Used to determine if finalize regeneration is needed at workflow completion
    script_gen_had_incomplete_actions: bool = False

    def __repr__(self) -> str:
        return f"SkyvernContext(request_id={self.request_id}, organization_id={self.organization_id}, task_id={self.task_id}, step_id={self.step_id}, workflow_id={self.workflow_id}, workflow_run_id={self.workflow_run_id}, task_v2_id={self.task_v2_id}, max_steps_override={self.max_steps_override}, run_id={self.run_id})"

    def __str__(self) -> str:
        return self.__repr__()

    def pop_totp_code(self, task_id: str) -> None:
        if task_id in self.totp_codes:
            self.totp_codes.pop(task_id)

    def add_magic_link_page(self, task_id: str, page: Page) -> None:
        self.magic_link_pages[task_id] = page

    def has_magic_link_page(self, task_id: str) -> bool:
        if task_id not in self.magic_link_pages:
            return False

        page = self.magic_link_pages[task_id]
        if page.is_closed():
            self.magic_link_pages.pop(task_id)
            return False
        return True


_context: ContextVar[SkyvernContext | None] = ContextVar(
    "Global context",
    default=None,
)


def current() -> SkyvernContext | None:
    """
    Get the current context

    Returns:
        The current context, or None if there is none
    """
    return _context.get()


def ensure_context() -> SkyvernContext:
    """
    Get the current context, or raise an error if there is none

    Returns:
        The current context if there is one

    Raises:
        RuntimeError: If there is no current context
    """
    context = current()
    if context is None:
        raise RuntimeError("No skyvern context")
    return context


def set(context: SkyvernContext) -> None:
    """
    Set the current context

    Args:
        context: The context to set

    Returns:
        None
    """
    _context.set(context)


def reset() -> None:
    """
    Reset the current context

    Returns:
        None
    """
    _context.set(None)



import os
from pathlib import Path

import structlog
from dotenv import set_key

from skyvern.config import settings
from skyvern.forge import app
from skyvern.forge.sdk.core import security
from skyvern.forge.sdk.schemas.organizations import Organization, OrganizationAuthTokenType
from skyvern.forge.sdk.services.org_auth_token_service import API_KEY_LIFETIME
from skyvern.utils.env_paths import resolve_backend_env_path, resolve_frontend_env_path

LOG = structlog.get_logger()
SKYVERN_LOCAL_ORG = "Skyvern-local"
SKYVERN_LOCAL_DOMAIN = "skyvern.local"


def _write_env(path: Path, key: str, value: str) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    if not path.exists():
        path.touch()
    set_key(path, key, value)
    LOG.info(".env written", path=str(path), key=key)


def fingerprint_token(value: str) -> str:
    return f"{value[:6]}…{value[-4:]}" if len(value) > 12 else "[redacted -- token too short]"


async def ensure_local_org() -> Organization:
    """Ensure the local development organization exists and return it."""
    organization = await app.DATABASE.get_organization_by_domain(SKYVERN_LOCAL_DOMAIN)
    if organization:
        return organization

    return await app.DATABASE.create_organization(
        organization_name=SKYVERN_LOCAL_ORG,
        domain=SKYVERN_LOCAL_DOMAIN,
        max_steps_per_run=10,
        max_retries_per_step=3,
    )


async def regenerate_local_api_key() -> tuple[str, str, str, str | None]:
    """Create a fresh API key for the local organization and persist it to env files.

    Returns:
        tuple: (api_key, org_id, backend_env_path, frontend_env_path_or_none)
    """
    organization = await ensure_local_org()
    org_id = organization.organization_id

    await app.DATABASE.invalidate_org_auth_tokens(
        organization_id=org_id,
        token_type=OrganizationAuthTokenType.api,
    )

    api_key = security.create_access_token(org_id, expires_delta=API_KEY_LIFETIME)
    await app.DATABASE.create_org_auth_token(
        organization_id=org_id,
        token=api_key,
        token_type=OrganizationAuthTokenType.api,
    )

    backend_env_path = resolve_backend_env_path()
    _write_env(backend_env_path, "SKYVERN_API_KEY", api_key)

    frontend_env_path = resolve_frontend_env_path()
    if frontend_env_path:
        _write_env(frontend_env_path, "VITE_SKYVERN_API_KEY", api_key)
    else:
        LOG.warning("Frontend directory not found; skipping VITE_SKYVERN_API_KEY update")

    settings.SKYVERN_API_KEY = api_key
    os.environ["SKYVERN_API_KEY"] = api_key

    LOG.info(
        "Local API key regenerated",
        organization_id=org_id,
        fingerprint=fingerprint_token(api_key),
    )
    return api_key, org_id, str(backend_env_path), str(frontend_env_path) if frontend_env_path else None



"""Backward-compatible re-exports from skyvern.cli.core.

MCP tools import from here; the canonical implementations live in core/.
"""

from __future__ import annotations

from skyvern.cli.core.client import get_skyvern
from skyvern.cli.core.session_manager import (
    BrowserNotAvailableError,
    SessionState,
    browser_session,
    close_current_session,
    get_current_session,
    get_page,
    no_browser_error,
    resolve_browser,
    set_current_session,
)

__all__ = [
    "BrowserNotAvailableError",
    "SessionState",
    "browser_session",
    "close_current_session",
    "get_current_session",
    "get_page",
    "get_skyvern",
    "no_browser_error",
    "resolve_browser",
    "set_current_session",
]



from typing import TYPE_CHECKING, Any

from playwright.async_api import BrowserContext, Page

from skyvern.library.skyvern_browser_page import SkyvernBrowserPage

if TYPE_CHECKING:
    from skyvern.library.skyvern import Skyvern


class SkyvernBrowser(BrowserContext):
    """A browser context wrapper that creates Skyvern-enabled pages.

    This class extends Playwright BrowserContext and provides methods to create
    SkyvernBrowserPage instances that combine traditional browser automation with
    AI-powered task execution capabilities. It manages browser session state and
    enables persistent browser sessions across multiple pages.

    Example:
        ```python
            skyvern = Skyvern.local()
            browser = await skyvern.launch_local_browser()

            # Get or create the working page
            page = await browser.get_working_page()

            # Create a new page
            new_page = await browser.new_page()
        ```

    Attributes:
        _browser_context: The underlying Playwright BrowserContext.
        _browser_session_id: Optional session ID for persistent browser sessions.
        _browser_address: Optional address for remote browser connections.
    """

    def __init__(
        self,
        skyvern: "Skyvern",
        browser_context: BrowserContext,
        *,
        browser_session_id: str | None = None,
        browser_address: str | None = None,
    ):
        super().__init__(browser_context)
        self._skyvern = skyvern
        self._browser_context = browser_context
        self._browser_session_id = browser_session_id
        self._browser_address = browser_address

        self.workflow_run_id: None | str = None

    def __getattribute__(self, name: str) -> Any:
        browser_context = object.__getattribute__(self, "_browser_context")
        if hasattr(browser_context, name):
            for cls in type(self).__mro__:
                if cls is BrowserContext:
                    break
                if name in cls.__dict__:
                    return object.__getattribute__(self, name)
            return getattr(browser_context, name)

        return object.__getattribute__(self, name)

    @property
    def browser_session_id(self) -> str | None:
        return self._browser_session_id

    @property
    def browser_address(self) -> str | None:
        return self._browser_address

    @property
    def skyvern(self) -> "Skyvern":
        return self._skyvern

    async def get_working_page(self) -> SkyvernBrowserPage:
        """Get the most recent page or create a new one if none exists.

        This method returns the last page in the browser context, or creates a new page
        if the context has no pages. This is useful for continuing work on an existing
        page without creating unnecessary new tabs.

        Returns:
            SkyvernBrowserPage: The most recent page wrapped with Skyvern capabilities.
        """
        if self._browser_context.pages:
            page = self._browser_context.pages[-1]
        else:
            page = await self._browser_context.new_page()
        return await self._create_skyvern_page(page)

    async def new_page(self) -> SkyvernBrowserPage:
        """Create a new page (tab) in the browser context.

        This method always creates a new page, similar to opening a new tab in a browser.
        The new page will have both Playwright's standard API and Skyvern's AI capabilities.

        Returns:
            SkyvernBrowserPage: A new page wrapped with Skyvern capabilities.
        """
        page = await self._browser_context.new_page()
        return await self._create_skyvern_page(page)

    async def _create_skyvern_page(self, page: Page) -> SkyvernBrowserPage:
        return SkyvernBrowserPage(self, page)

    async def close(self, **kwargs: Any) -> None:
        """Close the browser and optionally close the browser session.

        This method closes the browser context. If the browser is associated with a
        cloud browser session (has a browser_session_id), it will also close the
        browser session via the API, marking it as completed.

        Args:
            **kwargs: Arguments passed to the underlying BrowserContext.close() method.

        Example:
            ```python
            browser = await skyvern.launch_cloud_browser()
            # ... use the browser ...
            await browser.close()  # Closes both browser and cloud session
            ```
        """
        await self._browser_context.close(**kwargs)

        if self._browser_session_id:
            await self._skyvern.close_browser_session(self._browser_session_id)



"""
Utility functions for PDF parsing with fallback support.

This module provides robust PDF parsing that tries pypdf first and falls back
to pdfplumber if pypdf fails, ensuring maximum compatibility with various PDF formats.
"""

import pdfplumber
import structlog
from pypdf import PdfReader

from skyvern.constants import MAX_FILE_PARSE_INPUT_TOKENS
from skyvern.exceptions import PDFParsingError
from skyvern.forge.sdk.utils.sanitization import sanitize_postgres_text
from skyvern.utils.token_counter import count_tokens

LOG = structlog.get_logger(__name__)


def extract_pdf_file(
    file_path: str,
    file_identifier: str | None = None,
    max_tokens: int = MAX_FILE_PARSE_INPUT_TOKENS,
) -> str:
    """
    Extract text from a PDF file with fallback support.

    This function attempts to parse the PDF using pypdf first. If that fails,
    it automatically falls back to pdfplumber. This provides robust handling
    of various PDF formats, including those with corrupted streams or non-standard
    formatting that may cause pypdf to fail.

    Args:
        file_path: Path to the PDF file to parse
        file_identifier: Optional identifier for logging (e.g., URL or filename).
                        If not provided, uses file_path.

    Returns:
        Extracted text from all pages of the PDF

    Raises:
        PDFParsingError: When both pypdf and pdfplumber fail to parse the PDF

    Example:
        >>> text = extract_pdf_file("/path/to/file.pdf", "document.pdf")
        >>> print(f"Extracted {len(text)} characters")
    """
    identifier = file_identifier or file_path

    # Try pypdf first
    try:
        reader = PdfReader(file_path)
        extracted_text = ""
        current_tokens = 0
        page_count = len(reader.pages)

        for i in range(page_count):
            page_text = reader.pages[i].extract_text() or ""
            if max_tokens:
                page_tokens = count_tokens(page_text)
                if current_tokens + page_tokens > max_tokens:
                    LOG.warning(
                        "PDF text exceeds token limit, truncating at page boundary",
                        file_identifier=identifier,
                        pages_included=i,
                        total_pages=page_count,
                        max_tokens=max_tokens,
                    )
                    break
                current_tokens += page_tokens
            extracted_text += page_text + "\n"

        # Sanitize text to remove characters that cannot be stored in PostgreSQL
        extracted_text = sanitize_postgres_text(extracted_text)

        LOG.info(
            "Successfully parsed PDF with pypdf",
            file_identifier=identifier,
            page_count=page_count,
            text_length=len(extracted_text),
        )
        return extracted_text

    except Exception as pypdf_error:
        LOG.warning(
            "Failed to parse PDF with pypdf, trying pdfplumber",
            file_identifier=identifier,
            error=str(pypdf_error),
            error_type=type(pypdf_error).__name__,
        )

        # Fallback to pdfplumber
        try:
            with pdfplumber.open(file_path) as pdf:
                extracted_text = ""
                current_tokens = 0
                page_count = len(pdf.pages)

                for i, page in enumerate(pdf.pages):
                    page_text = page.extract_text()
                    if page_text:
                        if max_tokens:
                            page_tokens = count_tokens(page_text)
                            if current_tokens + page_tokens > max_tokens:
                                LOG.warning(
                                    "PDF text exceeds token limit, truncating at page boundary",
                                    file_identifier=identifier,
                                    pages_included=i,
                                    total_pages=page_count,
                                    max_tokens=max_tokens,
                                )
                                break
                            current_tokens += page_tokens
                        extracted_text += page_text + "\n"

                # Sanitize text to remove characters that cannot be stored in PostgreSQL
                extracted_text = sanitize_postgres_text(extracted_text)

                LOG.info(
                    "Successfully parsed PDF with pdfplumber",
                    file_identifier=identifier,
                    page_count=page_count,
                    text_length=len(extracted_text),
                )
                return extracted_text

        except Exception as pdfplumber_error:
            LOG.error(
                "Failed to parse PDF with both pypdf and pdfplumber",
                file_identifier=identifier,
                pypdf_error=str(pypdf_error),
                pdfplumber_error=str(pdfplumber_error),
            )
            raise PDFParsingError(
                file_identifier=identifier,
                pypdf_error=str(pypdf_error),
                pdfplumber_error=str(pdfplumber_error),
            )


def validate_pdf_file(
    file_path: str,
    file_identifier: str | None = None,
) -> bool:
    """
    Validate that a file is a readable PDF.

    This function attempts to validate the PDF using pypdf first. If that fails,
    it automatically falls back to pdfplumber validation.

    Args:
        file_path: Path to the PDF file to validate
        file_identifier: Optional identifier for logging (e.g., URL or filename).
                        If not provided, uses file_path.

    Returns:
        True if the PDF can be opened and read by at least one parser

    Raises:
        PDFParsingError: When both pypdf and pdfplumber fail to validate the PDF

    Example:
        >>> if validate_pdf_file("/path/to/file.pdf"):
        ...     print("Valid PDF file")
    """
    identifier = file_identifier or file_path

    # Try pypdf first
    try:
        reader = PdfReader(file_path)
        # Just check if we can access pages, don't read content yet
        _ = len(reader.pages)
        LOG.debug(
            "PDF validation successful with pypdf",
            file_identifier=identifier,
        )
        return True

    except Exception as pypdf_error:
        LOG.debug(
            "PDF validation with pypdf failed, trying pdfplumber",
            file_identifier=identifier,
            error=str(pypdf_error),
        )

        # Fallback to pdfplumber
        try:
            with pdfplumber.open(file_path) as pdf:
                _ = len(pdf.pages)

            LOG.info(
                "PDF validation: pypdf failed but pdfplumber succeeded",
                file_identifier=identifier,
                pypdf_error=str(pypdf_error),
            )
            return True

        except Exception as pdfplumber_error:
            LOG.error(
                "PDF validation failed with both pypdf and pdfplumber",
                file_identifier=identifier,
                pypdf_error=str(pypdf_error),
                pdfplumber_error=str(pdfplumber_error),
            )
            raise PDFParsingError(
                file_identifier=identifier,
                pypdf_error=str(pypdf_error),
                pdfplumber_error=str(pdfplumber_error),
            )



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class PromptAction(UniversalBaseModel):
    """
    Prompt action parameters.
    """

    prompt: str = pydantic.Field()
    """
    The prompt to send to the LLM
    """

    response_schema: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = pydantic.Field(default=None)
    """
    Optional JSON schema to structure the response
    """

    model: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = pydantic.Field(default=None)
    """
    Optional model configuration
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import asyncio
from typing import Any, Awaitable, Callable, Protocol

from fastapi import Request
from pydantic import BaseModel
from sse_starlette import EventSourceResponse, JSONServerSentEvent, ServerSentEvent

DEFAULT_KEEPALIVE_INTERVAL_SECONDS = 10


class EventSourceStream(Protocol):
    """Protocol for Server-Sent Events (SSE) streams."""

    async def send(self, data: Any) -> bool:
        """
        Send data as an SSE event.

        Returns:
            True if the event was queued successfully, False if disconnected or closed.
        """
        ...

    async def is_disconnected(self) -> bool:
        """Check if the client has disconnected."""
        ...

    async def close(self) -> None:
        """Signal that the stream is complete."""
        ...


class FastAPIEventSourceStream:
    """
    FastAPI implementation of EventSourceStream.

    This class provides a cleaner interface for sending SSE updates from async functions
    instead of using yield-based generators directly.

    Usage:
        @app.post("/stream")
        async def my_endpoint(request: Request) -> EventSourceResponse:
            async def handler(stream: EventSourceStream) -> None:
                await stream.send(MyUpdateModel(status="Processing..."))
                result = await do_work()
                await stream.send({"status": "Done", "result": result})

            return FastAPIEventSourceStream.create(request, handler)
    """

    def __init__(self, request: Request) -> None:
        self._request = request
        self._queue: asyncio.Queue[Any] = asyncio.Queue()
        self._closed = False

    async def send(self, data: Any) -> bool:
        """
        Send data as an SSE event. Accepts Pydantic models or dicts.

        Returns:
            True if the event was queued successfully, False if disconnected or closed.
        """
        if self._closed or await self.is_disconnected():
            return False
        await self._queue.put(data)
        return True

    async def is_disconnected(self) -> bool:
        """Check if the client has disconnected."""
        return await self._request.is_disconnected()

    async def close(self) -> None:
        """Signal that the stream is complete."""
        self._closed = True
        await self._queue.put(None)

    def _serialize(self, data: Any) -> Any:
        """Serialize data to JSON-compatible format."""
        if isinstance(data, BaseModel):
            return data.model_dump(mode="json")
        return data

    async def _generate(self) -> Any:
        """Internal generator that yields SSE events from the queue."""
        while True:
            try:
                data = await self._queue.get()
                if data is None:
                    break
                if await self.is_disconnected():
                    break
                yield JSONServerSentEvent(data=self._serialize(data))
            except Exception:
                break

    @classmethod
    def create(
        cls,
        request: Request,
        handler: Callable[[EventSourceStream], Awaitable[None]],
        ping_interval: int = DEFAULT_KEEPALIVE_INTERVAL_SECONDS,
    ) -> EventSourceResponse:
        """
        Create an EventSourceResponse that runs the handler with an EventSourceStream.

        Args:
            request: The FastAPI request object
            handler: An async function that receives the stream and sends events
            ping_interval: Interval in seconds for keep-alive pings (default: 10)

        Returns:
            An EventSourceResponse that can be returned from a FastAPI endpoint
        """
        stream = cls(request)

        async def event_generator() -> Any:
            task = asyncio.create_task(cls._run_handler(stream, handler))
            try:
                async for event in stream._generate():
                    yield event
            finally:
                if not task.done():
                    task.cancel()
                    try:
                        await task
                    except asyncio.CancelledError:
                        pass

        def ping_message_factory() -> ServerSentEvent:
            return ServerSentEvent(comment="keep-alive")

        return EventSourceResponse(
            event_generator(),
            ping=ping_interval,
            ping_message_factory=ping_message_factory,
        )

    @staticmethod
    async def _run_handler(
        stream: EventSourceStream,
        handler: Callable[[EventSourceStream], Awaitable[None]],
    ) -> None:
        """Run the handler and ensure the stream is closed when done."""
        try:
            await handler(stream)
        finally:
            await stream.close()



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .input_text_action_data import InputTextActionData


class InputTextAction(UniversalBaseModel):
    """
    Input text action parameters.
    """

    selector: typing.Optional[str] = pydantic.Field(default=None)
    """
    CSS selector for the element
    """

    value: typing.Optional[str] = pydantic.Field(default=None)
    """
    Value to input
    """

    intention: typing.Optional[str] = pydantic.Field(default=None)
    """
    The intention or goal of the input
    """

    data: typing.Optional[InputTextActionData] = pydantic.Field(default=None)
    """
    Additional context data
    """

    totp_identifier: typing.Optional[str] = pydantic.Field(default=None)
    """
    TOTP identifier for input_text actions
    """

    totp_url: typing.Optional[str] = pydantic.Field(default=None)
    """
    TOTP URL for input_text actions
    """

    timeout: typing.Optional[float] = pydantic.Field(default=None)
    """
    Timeout in milliseconds
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""
Contains registries for coordinating active WS connections (aka "channels", see
`./channels/README.md`).

NOTE: in AWS we had to turn on what amounts to sticky sessions for frontend apps,
so that an individual frontend app instance is guaranteed to always connect to
the same backend api instance. This is beccause the two registries here are
tied together via a `client_id` string.

The tale-of-the-tape is this:
  - frontend app requires two different channels (WS connections) to the backend api
    - one dedicated to streaming VNC's RFB protocol
    - the other dedicated to messaging (JSON)
  - both of these channels are stateful and need to coordinate with one another
"""

from __future__ import annotations

import typing as t

import structlog

if t.TYPE_CHECKING:
    from skyvern.forge.sdk.routes.streaming.channels.message import MessageChannel
    from skyvern.forge.sdk.routes.streaming.channels.vnc import VncChannel

LOG = structlog.get_logger()


# a registry for VNC channels, keyed by `client_id`
vnc_channels: dict[str, VncChannel] = {}


def add_vnc_channel(vnc_channel: VncChannel) -> None:
    vnc_channels[vnc_channel.client_id] = vnc_channel


def get_vnc_channel(client_id: str) -> t.Union[VncChannel, None]:
    return vnc_channels.get(client_id, None)


def del_vnc_channel(client_id: str) -> None:
    try:
        del vnc_channels[client_id]
    except KeyError:
        pass


# a registry for message channels, keyed by `client_id`
message_channels: dict[str, MessageChannel] = {}


def add_message_channel(message_channel: MessageChannel) -> None:
    message_channels[message_channel.client_id] = message_channel


def get_message_channel(client_id: str) -> t.Union[MessageChannel, None]:
    candidate = message_channels.get(client_id, None)

    if candidate and candidate.is_open:
        return candidate

    if candidate:
        LOG.info(
            "MessageChannel: message channel is not open; deleting it",
            client_id=candidate.client_id,
        )

        del_message_channel(candidate.client_id)

    return None


def del_message_channel(client_id: str) -> None:
    try:
        del message_channels[client_id]
    except KeyError:
        pass



import asyncio
import contextvars
from contextlib import asynccontextmanager
from functools import wraps
from typing import Any, AsyncContextManager, AsyncIterator, Callable

import structlog
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession, async_sessionmaker

LOG = structlog.get_logger()


def read_retry(retries: int = 3) -> Callable:
    """Decorator to retry async database operations on transient failures.

    Args:
        retries: Maximum number of retry attempts (default: 3)
    """

    def decorator(fn: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(fn)
        async def wrapper(
            base_db: "BaseAlchemyDB",
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            for attempt in range(retries):
                try:
                    return await fn(base_db, *args, **kwargs)
                except SQLAlchemyError as e:
                    if not base_db.is_retryable_error(e):
                        LOG.error("SQLAlchemyError", exc_info=True, attempt=attempt)
                        raise
                    if attempt >= retries - 1:
                        LOG.error("SQLAlchemyError after all retries", exc_info=True, attempt=attempt)
                        raise

                    backoff_time = 0.2 * (2**attempt)
                    LOG.warning(
                        "SQLAlchemyError retrying",
                        attempt=attempt,
                        backoff_time=backoff_time,
                        exc_info=True,
                    )
                    await asyncio.sleep(backoff_time)

                except Exception:
                    LOG.error("UnexpectedError", exc_info=True)
                    raise

            raise RuntimeError(f"Retry logic error in {fn.__name__}")

        return wrapper

    return decorator


class BaseAlchemyDB:
    """Base database client with connection and session management."""

    def __init__(self, db_engine: AsyncEngine) -> None:
        self.engine = db_engine
        self.Session = _SessionFactory(self, async_sessionmaker(bind=db_engine))

    def is_retryable_error(self, error: SQLAlchemyError) -> bool:
        """Check if a database error is retryable. Override in subclasses for specific error handling."""
        return False


class _SessionFactory:
    def __init__(self, db: BaseAlchemyDB, sessionmaker: async_sessionmaker[AsyncSession]) -> None:
        self._db = db
        self._sessionmaker = sessionmaker
        self._session_ctx: contextvars.ContextVar[AsyncSession | None] = contextvars.ContextVar(
            "skyvern_db_session",
            default=None,
        )

    def __call__(self) -> AsyncContextManager[AsyncSession]:
        return self._session()

    def __getattr__(self, name: str) -> Any:
        return getattr(self._sessionmaker, name)

    @asynccontextmanager
    async def _session(self) -> AsyncIterator[AsyncSession]:
        existing_session = self._session_ctx.get()
        if existing_session is not None:
            yield existing_session
            return

        session = self._sessionmaker()
        token = self._session_ctx.set(session)
        try:
            yield session
        finally:
            self._session_ctx.reset(token)
            try:
                await session.close()
            except SQLAlchemyError as e:
                # Handle transient errors during session cleanup gracefully.
                # This can happen on replicas when the connection is terminated due to
                # WAL replay conflicts. Since the actual DB operation already completed
                # successfully (we're in finally block cleanup), we just log and continue.
                if self._db.is_retryable_error(e):
                    LOG.warning(
                        "Transient error during session close (suppressed)",
                        error=str(e),
                    )
                else:
                    raise



"""add vnc support in debug session model

Revision ID: d15d85a702a4
Revises: bbcc08ba09f5
Create Date: 2025-09-12 13:13:37.428674+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "d15d85a702a4"
down_revision: Union[str, None] = "bbcc08ba09f5"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "debug_sessions",
        sa.Column("vnc_streaming_supported", sa.Boolean(), server_default=sa.text("true"), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("debug_sessions", "vnc_streaming_supported")
    # ### end Alembic commands ###



from typing import Any

import structlog
from fastapi import Depends, HTTPException, status

from skyvern.core.script_generations.real_skyvern_page_ai import RealSkyvernPageAi
from skyvern.core.script_generations.script_skyvern_page import ScriptSkyvernPage
from skyvern.exceptions import ScrapingFailed
from skyvern.forge import app
from skyvern.forge.sdk.api.files import validate_download_url
from skyvern.forge.sdk.core import skyvern_context
from skyvern.forge.sdk.core.skyvern_context import SkyvernContext
from skyvern.forge.sdk.routes.routers import base_router
from skyvern.forge.sdk.schemas.organizations import Organization
from skyvern.forge.sdk.schemas.sdk_actions import (
    RunSdkActionRequest,
    RunSdkActionResponse,
)
from skyvern.forge.sdk.schemas.tasks import TaskStatus
from skyvern.forge.sdk.services import org_auth_service
from skyvern.forge.sdk.workflow.models.workflow import (
    WorkflowRequestBody,
    WorkflowRunStatus,
)
from skyvern.schemas.workflows import BlockType, WorkflowStatus

LOG = structlog.get_logger()


@base_router.post(
    "/sdk/run_action",
    response_model=RunSdkActionResponse,
    summary="Run an SDK action",
    description="Execute a single SDK action with the specified parameters",
    tags=["SDK"],
    openapi_extra={
        "x-fern-sdk-method-name": "run_sdk_action",
    },
)
@base_router.post("/sdk/run_action/", include_in_schema=False)
async def run_sdk_action(
    action_request: RunSdkActionRequest,
    organization: Organization = Depends(org_auth_service.get_current_org),
) -> RunSdkActionResponse:
    """Execute a single SDK action with the specified parameters."""
    LOG.info(
        "Running SDK action",
        organization_id=organization.organization_id,
        action_type=action_request.action.type,
    )

    organization_id = organization.organization_id
    browser_session_id = action_request.browser_session_id
    browser_address = action_request.browser_address
    action = action_request.action

    # Use existing workflow_run_id if provided, otherwise create a new one
    if action_request.workflow_run_id:
        workflow_run = await app.DATABASE.get_workflow_run(
            workflow_run_id=action_request.workflow_run_id,
            organization_id=organization_id,
        )
        if not workflow_run:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Workflow run {action_request.workflow_run_id} not found",
            )
        workflow = await app.DATABASE.get_workflow(
            workflow_id=workflow_run.workflow_id,
            organization_id=organization_id,
        )
        if not workflow:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"Workflow {workflow_run.workflow_id} not found",
            )
    else:
        workflow = await app.WORKFLOW_SERVICE.create_empty_workflow(
            organization,
            title="SDK Workflow",
            status=WorkflowStatus.auto_generated,
        )
        workflow_run = await app.WORKFLOW_SERVICE.setup_workflow_run(
            request_id=None,
            workflow_request=WorkflowRequestBody(
                browser_session_id=browser_session_id,
                browser_address=browser_address,
            ),
            workflow_permanent_id=workflow.workflow_permanent_id,
            organization=organization,
            version=None,
        )
        workflow_run = await app.DATABASE.update_workflow_run(
            workflow_run_id=workflow_run.workflow_run_id,
            status=WorkflowRunStatus.completed,
        )

    task = await app.DATABASE.create_task(
        organization_id=organization_id,
        url=action_request.url,
        navigation_goal=action.get_navigation_goal(),
        navigation_payload=action.get_navigation_payload(),
        data_extraction_goal=None,
        title=f"SDK Action Task: {action_request.action.type}",
        workflow_run_id=workflow_run.workflow_run_id,
        browser_session_id=browser_session_id,
        browser_address=browser_address,
    )

    step = await app.DATABASE.create_step(
        task.task_id,
        order=0,
        retry_index=0,
        organization_id=organization.organization_id,
    )

    await app.DATABASE.create_workflow_run_block(
        workflow_run_id=workflow_run.workflow_run_id,
        organization_id=organization_id,
        block_type=BlockType.ACTION,
        task_id=task.task_id,
    )

    await app.WORKFLOW_CONTEXT_MANAGER.initialize_workflow_run_context(
        organization,
        workflow_run.workflow_run_id,
        workflow.title,
        workflow.workflow_id,
        workflow.workflow_permanent_id,
        [],
        [],
        [],
        [],
        None,
        workflow,
    )

    context = skyvern_context.ensure_context()
    skyvern_context.set(
        SkyvernContext(
            request_id=context.request_id,
            organization_id=task.organization_id,
            task_id=task.task_id,
            step_id=step.step_id,
            browser_session_id=browser_session_id,
            max_screenshot_scrolls=task.max_screenshot_scrolls,
            workflow_id=workflow.workflow_id,
            workflow_run_id=workflow_run.workflow_run_id,
        )
    )
    result: Any | None = None
    try:
        scraped_page = await ScriptSkyvernPage.create_scraped_page(browser_session_id=browser_session_id)
        page = await scraped_page._browser_state.must_get_working_page()
        page_ai = RealSkyvernPageAi(scraped_page, page)

        if action.type == "ai_click":
            result = await page_ai.ai_click(
                selector=action.selector,
                intention=action.intention,
                data=action.data,
                timeout=action.timeout,
            )
        elif action.type == "ai_input_text":
            result = await page_ai.ai_input_text(
                selector=action.selector,
                value=action.value,
                intention=action.intention,
                data=action.data,
                totp_identifier=action.totp_identifier,
                totp_url=action.totp_url,
                timeout=action.timeout,
            )
        elif action.type == "ai_select_option":
            result = await page_ai.ai_select_option(
                selector=action.selector,
                value=action.value,
                intention=action.intention,
                data=action.data,
                timeout=action.timeout,
            )
        elif action.type == "ai_upload_file":
            if action.file_url and not validate_download_url(action.file_url):
                raise HTTPException(status_code=400, detail="Unsupported file url")
            result = await page_ai.ai_upload_file(
                selector=action.selector,
                files=action.file_url,
                intention=action.intention,
                data=action.data,
                timeout=action.timeout,
            )
        elif action.type == "ai_act":
            await page_ai.ai_act(
                prompt=action.intention,
            )
            result = None
        elif action.type == "extract":
            extract_result = await page_ai.ai_extract(
                prompt=action.prompt,
                schema=action.extract_schema,
                error_code_mapping=action.error_code_mapping,
                intention=action.intention,
                data=action.data,
            )
            result = extract_result
        elif action.type == "locate_element":
            xpath_result = await page_ai.ai_locate_element(
                prompt=action.prompt,
            )
            result = xpath_result
        elif action.type == "validate":
            validation_result = await page_ai.ai_validate(
                prompt=action.prompt,
                model=action.model,
            )
            result = validation_result
        elif action.type == "prompt":
            prompt_result = await page_ai.ai_prompt(
                prompt=action.prompt,
                schema=action.response_schema,
                model=action.model,
            )
            result = prompt_result
        await app.DATABASE.update_task(
            task_id=task.task_id,
            organization_id=organization_id,
            status=TaskStatus.completed,
        )
    except ScrapingFailed as e:
        await app.DATABASE.update_task(
            task_id=task.task_id,
            organization_id=organization_id,
            status=TaskStatus.failed,
            failure_reason=str(e.reason) if e.reason else str(e),
        )
        LOG.warning(
            "SDK action failed due to scraping error",
            action_type=action.type,
            error=str(e),
        )
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=e.reason or str(e))
    except Exception as e:
        await app.DATABASE.update_task(
            task_id=task.task_id,
            organization_id=organization_id,
            status=TaskStatus.failed,
            failure_reason=str(e),
        )
        LOG.error(
            "SDK action failed",
            action_type=action.type,
            error=str(e),
            exc_info=True,
        )
        raise
    finally:
        skyvern_context.reset()

    return RunSdkActionResponse(
        workflow_run_id=workflow_run.workflow_run_id,
        result=result,
    )



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .file_info import FileInfo
from .run_status import RunStatus
from .script_run_response import ScriptRunResponse
from .task_run_request import TaskRunRequest
from .task_run_response_output import TaskRunResponseOutput
from .workflow_run_request import WorkflowRunRequest
from .workflow_run_response_output import WorkflowRunResponseOutput


class GetRunResponse_TaskV1(UniversalBaseModel):
    run_type: typing.Literal["task_v1"] = "task_v1"
    run_id: str
    status: RunStatus
    output: typing.Optional[TaskRunResponseOutput] = None
    downloaded_files: typing.Optional[typing.List[FileInfo]] = None
    recording_url: typing.Optional[str] = None
    screenshot_urls: typing.Optional[typing.List[str]] = None
    failure_reason: typing.Optional[str] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    queued_at: typing.Optional[dt.datetime] = None
    started_at: typing.Optional[dt.datetime] = None
    finished_at: typing.Optional[dt.datetime] = None
    app_url: typing.Optional[str] = None
    browser_session_id: typing.Optional[str] = None
    browser_profile_id: typing.Optional[str] = None
    max_screenshot_scrolls: typing.Optional[int] = None
    script_run: typing.Optional[ScriptRunResponse] = None
    errors: typing.Optional[typing.List[typing.Dict[str, typing.Optional[typing.Any]]]] = None
    step_count: typing.Optional[int] = None
    run_request: typing.Optional[TaskRunRequest] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class GetRunResponse_TaskV2(UniversalBaseModel):
    run_type: typing.Literal["task_v2"] = "task_v2"
    run_id: str
    status: RunStatus
    output: typing.Optional[TaskRunResponseOutput] = None
    downloaded_files: typing.Optional[typing.List[FileInfo]] = None
    recording_url: typing.Optional[str] = None
    screenshot_urls: typing.Optional[typing.List[str]] = None
    failure_reason: typing.Optional[str] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    queued_at: typing.Optional[dt.datetime] = None
    started_at: typing.Optional[dt.datetime] = None
    finished_at: typing.Optional[dt.datetime] = None
    app_url: typing.Optional[str] = None
    browser_session_id: typing.Optional[str] = None
    browser_profile_id: typing.Optional[str] = None
    max_screenshot_scrolls: typing.Optional[int] = None
    script_run: typing.Optional[ScriptRunResponse] = None
    errors: typing.Optional[typing.List[typing.Dict[str, typing.Optional[typing.Any]]]] = None
    step_count: typing.Optional[int] = None
    run_request: typing.Optional[TaskRunRequest] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class GetRunResponse_OpenaiCua(UniversalBaseModel):
    run_type: typing.Literal["openai_cua"] = "openai_cua"
    run_id: str
    status: RunStatus
    output: typing.Optional[TaskRunResponseOutput] = None
    downloaded_files: typing.Optional[typing.List[FileInfo]] = None
    recording_url: typing.Optional[str] = None
    screenshot_urls: typing.Optional[typing.List[str]] = None
    failure_reason: typing.Optional[str] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    queued_at: typing.Optional[dt.datetime] = None
    started_at: typing.Optional[dt.datetime] = None
    finished_at: typing.Optional[dt.datetime] = None
    app_url: typing.Optional[str] = None
    browser_session_id: typing.Optional[str] = None
    browser_profile_id: typing.Optional[str] = None
    max_screenshot_scrolls: typing.Optional[int] = None
    script_run: typing.Optional[ScriptRunResponse] = None
    errors: typing.Optional[typing.List[typing.Dict[str, typing.Optional[typing.Any]]]] = None
    step_count: typing.Optional[int] = None
    run_request: typing.Optional[TaskRunRequest] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class GetRunResponse_AnthropicCua(UniversalBaseModel):
    run_type: typing.Literal["anthropic_cua"] = "anthropic_cua"
    run_id: str
    status: RunStatus
    output: typing.Optional[TaskRunResponseOutput] = None
    downloaded_files: typing.Optional[typing.List[FileInfo]] = None
    recording_url: typing.Optional[str] = None
    screenshot_urls: typing.Optional[typing.List[str]] = None
    failure_reason: typing.Optional[str] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    queued_at: typing.Optional[dt.datetime] = None
    started_at: typing.Optional[dt.datetime] = None
    finished_at: typing.Optional[dt.datetime] = None
    app_url: typing.Optional[str] = None
    browser_session_id: typing.Optional[str] = None
    browser_profile_id: typing.Optional[str] = None
    max_screenshot_scrolls: typing.Optional[int] = None
    script_run: typing.Optional[ScriptRunResponse] = None
    errors: typing.Optional[typing.List[typing.Dict[str, typing.Optional[typing.Any]]]] = None
    step_count: typing.Optional[int] = None
    run_request: typing.Optional[TaskRunRequest] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class GetRunResponse_UiTars(UniversalBaseModel):
    run_type: typing.Literal["ui_tars"] = "ui_tars"
    run_id: str
    status: RunStatus
    output: typing.Optional[TaskRunResponseOutput] = None
    downloaded_files: typing.Optional[typing.List[FileInfo]] = None
    recording_url: typing.Optional[str] = None
    screenshot_urls: typing.Optional[typing.List[str]] = None
    failure_reason: typing.Optional[str] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    queued_at: typing.Optional[dt.datetime] = None
    started_at: typing.Optional[dt.datetime] = None
    finished_at: typing.Optional[dt.datetime] = None
    app_url: typing.Optional[str] = None
    browser_session_id: typing.Optional[str] = None
    browser_profile_id: typing.Optional[str] = None
    max_screenshot_scrolls: typing.Optional[int] = None
    script_run: typing.Optional[ScriptRunResponse] = None
    errors: typing.Optional[typing.List[typing.Dict[str, typing.Optional[typing.Any]]]] = None
    step_count: typing.Optional[int] = None
    run_request: typing.Optional[TaskRunRequest] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


class GetRunResponse_WorkflowRun(UniversalBaseModel):
    run_type: typing.Literal["workflow_run"] = "workflow_run"
    run_id: str
    status: RunStatus
    output: typing.Optional[WorkflowRunResponseOutput] = None
    downloaded_files: typing.Optional[typing.List[FileInfo]] = None
    recording_url: typing.Optional[str] = None
    screenshot_urls: typing.Optional[typing.List[str]] = None
    failure_reason: typing.Optional[str] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    queued_at: typing.Optional[dt.datetime] = None
    started_at: typing.Optional[dt.datetime] = None
    finished_at: typing.Optional[dt.datetime] = None
    app_url: typing.Optional[str] = None
    browser_session_id: typing.Optional[str] = None
    browser_profile_id: typing.Optional[str] = None
    max_screenshot_scrolls: typing.Optional[int] = None
    script_run: typing.Optional[ScriptRunResponse] = None
    errors: typing.Optional[typing.List[typing.Dict[str, typing.Optional[typing.Any]]]] = None
    step_count: typing.Optional[int] = None
    run_with: typing.Optional[str] = None
    ai_fallback: typing.Optional[bool] = None
    run_request: typing.Optional[WorkflowRunRequest] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


GetRunResponse = typing.Union[
    GetRunResponse_TaskV1,
    GetRunResponse_TaskV2,
    GetRunResponse_OpenaiCua,
    GetRunResponse_AnthropicCua,
    GetRunResponse_UiTars,
    GetRunResponse_WorkflowRun,
]



import asyncio
import base64
import json
import pathlib
import typing as t
import zlib

import structlog

import skyvern.services.browser_recording.state_machines as sm
from skyvern.client.types.workflow_definition_yaml_blocks_item import (
    WorkflowDefinitionYamlBlocksItem_Action,
    WorkflowDefinitionYamlBlocksItem_GotoUrl,
    WorkflowDefinitionYamlBlocksItem_Wait,
)
from skyvern.client.types.workflow_definition_yaml_parameters_item import WorkflowDefinitionYamlParametersItem_Workflow
from skyvern.forge import app
from skyvern.forge.prompts import prompt_engine
from skyvern.services.browser_recording.types import (
    Action,
    ActionBlockable,
    ActionKind,
    ActionUrlChange,
    ActionWait,
    ExfiltratedCdpEvent,
    ExfiltratedConsoleEvent,
    ExfiltratedEvent,
    OutputBlock,
)

LOG = structlog.get_logger(__name__)

# avoid decompression bombs
MAX_BASE64_SIZE = 14 * 1024 * 1024  # ~10MB compressed + base64 overhead


class Processor:
    """
    Process browser session recordings into workflow definition blocks.
    """

    def __init__(
        self,
        browser_session_id: str,
        organization_id: str,
        workflow_permanent_id: str,
    ) -> None:
        self.browser_session_id = browser_session_id
        self.organization_id = organization_id
        self.workflow_permanent_id = workflow_permanent_id

    @property
    def class_name(self) -> str:
        return self.__class__.__name__

    @property
    def identity(self) -> dict[str, str]:
        return dict(
            browser_session_id=self.browser_session_id,
            organization_id=self.organization_id,
            workflow_permanent_id=self.workflow_permanent_id,
        )

    def decompress(self, base64_payload: str) -> bytes | None:
        """
        Decode a base64 string, decompress it using gzip, and return it.
        """

        if len(base64_payload) > MAX_BASE64_SIZE:
            LOG.warning(f"{self.class_name}: base64 payload too large: {len(base64_payload)} bytes", **self.identity)
            return None

        try:
            # base64 decode -> gzip binary data
            #
            # NOTE(llm): The data sent from btoa() is technically a "non-standard"
            # Base64, but Python's standard decoder is usually robust enough to
            # handle it.
            compressed_data: bytes = base64.b64decode(base64_payload)
        except Exception as ex:
            LOG.warning(f"{self.class_name} failed to decode Base64 payload", exc_info=ex, **self.identity)
            return None

        try:
            # gzip decompression -> bytes
            #
            # NOTE(llm): We use zlib.decompress with wbits=16 + zlib.MAX_WBITS (31).
            # This tells zlib to automatically detect and handle Gzip headers,
            # which is essential since the browser used CompressionStream('gzip').
            # Using zlib is often faster than the higher-level gzip module for this
            # purpose.
            decompressed_bytes: bytes = zlib.decompress(compressed_data, wbits=16 + zlib.MAX_WBITS)
        except zlib.error as e:
            LOG.warning(f"{self.class_name} decompression error: {e}", **self.identity)
            # Log the error, maybe log the first few characters of the payload for debugging
            return None

        return decompressed_bytes

    def serialize(self, decompressed_bytes: bytes | None) -> list[dict[str, t.Any]]:
        """
        Convert decompressed bytes into a list of events (Python list/dictionary).
        """
        if not decompressed_bytes:
            LOG.warning(f"{self.class_name} No decompressed bytes to serialize", **self.identity)
            return []

        try:
            # bytes -> JSON string
            json_string: str = decompressed_bytes.decode("utf-8")
        except Exception as e:
            LOG.warning(f"{self.class_name} decode error: {e}", **self.identity)
            return []

        try:
            # JSON string -> list of dicts
            events_list: list[dict[str, t.Any]] = json.loads(json_string)
        except Exception as e:
            LOG.warning(f"{self.class_name} JSON parsing error: {e}", **self.identity)
            return []

        if not isinstance(events_list, list):
            LOG.warning(f"{self.class_name} Expected a list of events, got:", type(events_list), **self.identity)
            return []

        return events_list

    def reify(self, events_list: list[dict[str, t.Any]]) -> list[ExfiltratedEvent]:
        """
        Convert a list of event dictionaries into a list of `ExfiltratedEvent`s.
        """

        if not events_list:
            LOG.warning(f"{self.class_name} No events to reify", **self.identity)
            return []

        reified_events: list[ExfiltratedEvent] = []
        for event in events_list:
            if event.get("source") == "cdp":
                try:
                    reified_event = ExfiltratedCdpEvent(**event)
                except Exception as e:
                    LOG.warning(f"{self.class_name} Failed to reify CDP event: {e}", **self.identity)
                    continue
            elif event.get("source") == "console":
                try:
                    reified_event = ExfiltratedConsoleEvent(**event)
                except Exception as e:
                    LOG.warning(f"{self.class_name} Failed to reify console event: {e}", **self.identity)
                    continue
            else:
                LOG.error(f"{self.class_name} Unknown event source: {event.get('source')}", **self.identity)
                continue
            reified_events.append(reified_event)

        return reified_events

    def compressed_chunks_to_events(self, compressed_chunks: list[str]) -> list[ExfiltratedEvent]:
        """
        Convert a list of base64 encoded and compressed (gzip) event strings into
        a list of `ExfiltratedEvent`s.
        """
        all_events: list[ExfiltratedEvent] = []

        for compressed_chunk in compressed_chunks:
            decompressed = self.decompress(compressed_chunk)
            serialized = self.serialize(decompressed)
            reified = self.reify(serialized)
            all_events.extend(reified)

        return all_events

    def events_to_actions(
        self,
        events: list[ExfiltratedEvent],
        machines: list[sm.StateMachine] | None = None,
    ) -> list[Action]:
        """
        Convert a list of `ExfiltratedEvent`s into `Action`s.
        """
        actions: list[Action] = []

        machines = machines or [
            sm.Click(),
            sm.Hover(),
            sm.InputText(),
            sm.UrlChange(),
            sm.Wait(),
        ]

        for event in events:
            for machine in machines:
                action = machine.tick(event, actions)

                if not action:
                    continue

                allow_action = True

                for m in machines:
                    if not m.on_action(action, actions):
                        allow_action = False
                        LOG.debug(
                            f"{self.class_name} action vetoed by state machine {m.__class__.__name__}",
                            action=action,
                            **self.identity,
                        )

                if allow_action:
                    actions.append(action)
                else:
                    # if an action was vetoed, we do not allow further processing
                    # of this event through subsequent state machines
                    break

        return actions

    def dedupe_block_labels(self, suspects: list[OutputBlock]) -> list[OutputBlock]:
        """
        Detect if any block labels are duplicated, and, if so, rename them for
        uniqueness.
        """

        blocks: list[OutputBlock] = []
        labels: set[str] = set()

        for block in suspects:
            if block.label not in labels:
                labels.add(block.label)
                blocks.append(block)
                continue
            else:
                original_label = block.label
                count = 0
                while True:
                    new_label = f"{original_label}_{count}"
                    if new_label not in labels:
                        cls = block.__class__
                        data = block.model_dump() | {"label": new_label}
                        new_block = cls(**data)
                        blocks.append(new_block)
                        labels.add(new_label)
                        break
                    count += 1

        return blocks

    async def actions_to_blocks(self, actions: list[Action]) -> list[OutputBlock]:
        """
        Convert a list of `Action` objects into workflow definition (YAML) blocks.
        """
        tasks: list[asyncio.Task] = []

        for action in actions:
            action_kind = action.kind.value

            match action.kind:
                case ActionKind.CLICK | ActionKind.HOVER | ActionKind.INPUT_TEXT:
                    task = asyncio.create_task(self.create_action_block(action))
                    tasks.append(task)
                case ActionKind.URL_CHANGE:
                    task = asyncio.create_task(self.create_url_block(action))
                    tasks.append(task)
                case ActionKind.WAIT:
                    task = asyncio.create_task(self.create_wait_block(action))
                    tasks.append(task)
                case _:
                    LOG.warning(
                        f"{self.class_name} Unknown action kind: {action_kind}",
                        action=action,
                        **self.identity,
                    )
                    continue

        blocks: list[OutputBlock] = await asyncio.gather(*tasks)

        blocks = self.dedupe_block_labels(blocks)

        return blocks

    def blocks_to_parameters(self, blocks: list[OutputBlock]) -> list[WorkflowDefinitionYamlParametersItem_Workflow]:
        """
        Convert a list of workflow definition (YAML) blocks into workflow definition (YAML) parameters.
        """
        parameter_names: set[str] = set()

        for block in blocks:
            if isinstance(block, WorkflowDefinitionYamlBlocksItem_Action):
                for param_name in block.parameter_keys or []:
                    parameter_names.add(param_name)

        parameters: list[WorkflowDefinitionYamlParametersItem_Workflow] = []

        for param_name in parameter_names:
            parameter = WorkflowDefinitionYamlParametersItem_Workflow(
                key=param_name,
                workflow_parameter_type="string",
                default_value="",
                description="",
            )
            parameters.append(parameter)

        return parameters

    async def create_action_block(self, action: ActionBlockable) -> WorkflowDefinitionYamlBlocksItem_Action:
        """
        Create a YAML action block from an `ActionBlockable`.
        """

        DEFAULT_BLOCK_TITLE = "Browser Action"

        if action.kind == ActionKind.INPUT_TEXT:
            prompt_name = "recording-action-block-prompt-input-text"
        else:
            prompt_name = "recording-action-block-prompt"

        metadata_prompt = prompt_engine.load_prompt(
            prompt_name,
            action=action,
        )

        metadata_response = await app.LLM_API_HANDLER(
            prompt=metadata_prompt,
            prompt_name=prompt_name,
            organization_id=self.organization_id,
        )

        block_label: str = metadata_response.get("block_label", None) or "act"
        title: str = metadata_response.get("title", None) or DEFAULT_BLOCK_TITLE
        navigation_goal: str = metadata_response.get("prompt", "")
        parameter_name: dict | None = metadata_response.get("parameter_name", None)

        block = WorkflowDefinitionYamlBlocksItem_Action(
            label=block_label,
            title=title,
            navigation_goal=navigation_goal,
            error_code_mapping=None,
            parameters=[parameter_name] if parameter_name else [],  # sic(jdo): the frontend requires this
            parameter_keys=[parameter_name.get("key")] if parameter_name else [],
        )

        return block

    async def create_url_block(self, action: ActionUrlChange) -> WorkflowDefinitionYamlBlocksItem_GotoUrl:
        """
        Create a YAML goto URL block from an `ActionUrlChange`.
        """

        prompt_name = "recording-go-to-url-block-prompt"

        metadata_prompt = prompt_engine.load_prompt(
            prompt_name,
            action=action,
        )

        metadata_response = await app.LLM_API_HANDLER(
            prompt=metadata_prompt,
            prompt_name=prompt_name,
            organization_id=self.organization_id,
        )

        block_label: str = metadata_response.get("block_label", None) or "goto_url"

        block = WorkflowDefinitionYamlBlocksItem_GotoUrl(
            label=block_label,
            url=action.url,
        )

        return block

    async def create_wait_block(self, action: ActionWait) -> WorkflowDefinitionYamlBlocksItem_Wait:
        """
        Create a YAML wait block from an `ActionWait`.
        """

        prompt_name = "recording-wait-block-prompt"

        metadata_prompt = prompt_engine.load_prompt(
            prompt_name,
            action=action,
        )

        metadata_response = await app.LLM_API_HANDLER(
            prompt=metadata_prompt,
            prompt_name=prompt_name,
            organization_id=self.organization_id,
        )

        block_label: str = metadata_response.get("block_label", None) or "wait"

        block = WorkflowDefinitionYamlBlocksItem_Wait(
            label=block_label,
            wait_sec=int(max(action.duration_ms / 1000.0, ActionWait.MIN_DURATION_THRESHOLD_MS / 1000.0)),
        )

        return block

    async def process(
        self, compressed_chunks: list[str]
    ) -> tuple[list[OutputBlock], list[WorkflowDefinitionYamlParametersItem_Workflow]]:
        """
        Process the compressed browser session recording into workflow definition blocks.
        """

        events = self.compressed_chunks_to_events(compressed_chunks)
        actions = self.events_to_actions(events)
        blocks = await self.actions_to_blocks(actions)
        parameters = self.blocks_to_parameters(blocks)

        return blocks, parameters


class BrowserSessionRecordingService:
    async def process_recording(
        self,
        browser_session_id: str,
        organization_id: str,
        workflow_permanent_id: str,
        compressed_chunks: list[str],
    ) -> tuple[list[OutputBlock], list[WorkflowDefinitionYamlParametersItem_Workflow]]:
        """
        Process compressed browser session recording events into workflow definition blocks.
        """
        processor = Processor(
            browser_session_id,
            organization_id,
            workflow_permanent_id,
        )

        return await processor.process(compressed_chunks)


async def smoke() -> None:
    with open(pathlib.Path("/path/to/uncompressed/events.json")) as f:
        raw_events: list[dict] = json.load(f)

    events: list[ExfiltratedEvent] = []

    for i, raw_event in enumerate(raw_events):
        if not isinstance(raw_event, dict):
            LOG.debug(f"~ skipping non-dict event: {raw_event}")
            continue
        if raw_event.get("source") == "cdp":
            try:
                event = ExfiltratedCdpEvent(**raw_event)
            except Exception:
                LOG.exception(f"{i} Failed to parse exfiltrated CDP event")
                LOG.debug(f"~ raw event: {json.dumps(raw_event, sort_keys=True, indent=2)}")
                continue
            events.append(event)
        elif raw_event.get("source") == "console":
            event = ExfiltratedConsoleEvent(**raw_event)
            events.append(event)

    LOG.debug(f"{len(events)} events.")

    my_local_org_id = "o_389844905020748346"
    processor = Processor("pbs_123", my_local_org_id, "wpid_123")
    actions = processor.events_to_actions(events)

    LOG.debug(f"{len(actions)} actions:")

    for action in actions:
        id = action.target.sky_id if action.target.sky_id else action.target.id
        text = ",".join(action.target.texts or [])
        LOG.debug(f"  {action.kind} [{id}] [{text}] @ {action.url}")

    blocks = await processor.actions_to_blocks(actions)

    LOG.debug(f"{len(blocks)} blocks:")

    for block in blocks:
        LOG.debug(f"  {block.label}")

        if isinstance(block, WorkflowDefinitionYamlBlocksItem_Action):
            LOG.debug(f"    title: {block.title}")
            LOG.debug(f"    nav goal: {block.navigation_goal}")

        if isinstance(block, WorkflowDefinitionYamlBlocksItem_GotoUrl):
            LOG.debug(f"    url: {block.url}")

        if isinstance(block, WorkflowDefinitionYamlBlocksItem_Wait):
            LOG.debug(f"    wait sec: {block.wait_sec}")


# if __name__ == "__main__":
#     from skyvern.forge.forge_app_initializer import start_forge_app

#     start_forge_app()

#     asyncio.run(smoke())



"""workflow copilot chat

Revision ID: db8667f8ce63
Revises: e393f33ec711
Create Date: 2026-01-06 21:48:28.396490+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "db8667f8ce63"
down_revision: Union[str, None] = "e393f33ec711"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "workflow_copilot_chat_messages",
        sa.Column("workflow_copilot_chat_message_id", sa.String(), nullable=False),
        sa.Column("workflow_copilot_chat_id", sa.String(), nullable=False),
        sa.Column("organization_id", sa.String(), nullable=False),
        sa.Column("sender", sa.String(), nullable=False),
        sa.Column("content", sa.UnicodeText(), nullable=False),
        sa.Column("global_llm_context", sa.UnicodeText(), nullable=True),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("modified_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("workflow_copilot_chat_message_id"),
    )
    op.create_index(
        op.f("ix_workflow_copilot_chat_messages_workflow_copilot_chat_id"),
        "workflow_copilot_chat_messages",
        ["workflow_copilot_chat_id"],
        unique=False,
    )
    op.create_table(
        "workflow_copilot_chats",
        sa.Column("workflow_copilot_chat_id", sa.String(), nullable=False),
        sa.Column("organization_id", sa.String(), nullable=False),
        sa.Column("workflow_permanent_id", sa.String(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("modified_at", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("workflow_copilot_chat_id"),
    )
    op.create_index(
        op.f("ix_workflow_copilot_chats_workflow_permanent_id"),
        "workflow_copilot_chats",
        ["workflow_permanent_id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_workflow_copilot_chats_workflow_permanent_id"), table_name="workflow_copilot_chats")
    op.drop_table("workflow_copilot_chats")
    op.drop_index(
        op.f("ix_workflow_copilot_chat_messages_workflow_copilot_chat_id"), table_name="workflow_copilot_chat_messages"
    )
    op.drop_table("workflow_copilot_chat_messages")
    # ### end Alembic commands ###



"""workflow.use_cache -> generate_script

Revision ID: 944ef972e5a8
Revises: a027553be970
Create Date: 2025-08-10 15:57:46.353038+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "944ef972e5a8"
down_revision: Union[str, None] = "a027553be970"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflows", sa.Column("generate_script", sa.Boolean(), nullable=False, server_default=sa.false()))

    # Copy data from use_cache to generate_script
    op.execute("UPDATE workflows SET generate_script = use_cache")

    # Remove server default after data migration
    op.alter_column("workflows", "generate_script", server_default=None)

    # Drop the old use_cache column
    op.drop_column("workflows", "use_cache")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    # Add back use_cache column
    op.add_column("workflows", sa.Column("use_cache", sa.Boolean(), nullable=False, server_default=sa.false()))

    # Copy data from generate_script back to use_cache
    op.execute("UPDATE workflows SET use_cache = generate_script")

    # Remove server default after data migration
    op.alter_column("workflows", "use_cache", server_default=None)

    # Drop the new generate_script column
    op.drop_column("workflows", "generate_script")
    # ### end Alembic commands ###



# This file was auto-generated by Fern from our API Definition.

import typing

ActionType = typing.Union[
    typing.Literal[
        "click",
        "input_text",
        "upload_file",
        "download_file",
        "select_option",
        "checkbox",
        "wait",
        "hover",
        "null_action",
        "solve_captcha",
        "terminate",
        "complete",
        "reload_page",
        "close_page",
        "extract",
        "verification_code",
        "goto_url",
        "scroll",
        "keypress",
        "move",
        "drag",
        "left_mouse",
    ],
    typing.Any,
]



"""
Channels (see ./channels/README.md) variously rely on the state of one of
these database entities:
  - browser session
  - task
  - workflow run

That is, channels are created on the basis of one of those entities, and that
entity must be in a valid state for the channel to continue.

So, in order to continue operating a channel, we need to periodically verify
that the underlying entity is still valid. This module provides logic to
perform those verifications.
"""

from __future__ import annotations

import asyncio
import typing as t
from datetime import datetime, timedelta

import structlog

from skyvern.config import settings
from skyvern.forge import app
from skyvern.forge.sdk.schemas.persistent_browser_sessions import AddressablePersistentBrowserSession, is_final_status
from skyvern.forge.sdk.schemas.tasks import Task, TaskStatus
from skyvern.forge.sdk.workflow.models.workflow import WorkflowRun, WorkflowRunStatus

if t.TYPE_CHECKING:
    from skyvern.forge.sdk.routes.streaming.channels.message import MessageChannel
    from skyvern.forge.sdk.routes.streaming.channels.vnc import VncChannel

LOG = structlog.get_logger()


class Constants:
    POLL_INTERVAL_FOR_VERIFICATION_SECONDS = 5


async def verify_browser_session(
    browser_session_id: str,
    organization_id: str,
) -> AddressablePersistentBrowserSession | None:
    """
    Verify the browser session exists, and is usable.
    """
    if settings.ENV == "local":
        dummy_browser_session = AddressablePersistentBrowserSession(
            persistent_browser_session_id=browser_session_id,
            organization_id=organization_id,
            browser_address="0.0.0.0:9223",
            ip_address="localhost",
            created_at=datetime.now(),
            modified_at=datetime.now(),
        )

        return dummy_browser_session

    browser_session = await app.PERSISTENT_SESSIONS_MANAGER.get_session(browser_session_id, organization_id)

    if not browser_session:
        LOG.info(
            "No browser session found.",
            browser_session_id=browser_session_id,
            organization_id=organization_id,
        )
        return None

    if is_final_status(browser_session.status):
        return None

    started_at = browser_session.started_at
    timeout_minutes = browser_session.timeout_minutes

    if started_at and timeout_minutes:
        current_time = datetime.utcnow()
        times_out_at = started_at + timedelta(minutes=timeout_minutes)

        if current_time > times_out_at:
            LOG.info(
                "Browser session invalid, as it has timed out, but is still in a non-final status. This is likely a bug!",
                browser_session_id=browser_session_id,
                organization_id=organization_id,
                timeout_minutes=timeout_minutes,
                started_at=started_at.isoformat(),
                now=current_time.isoformat(),
                times_out_at=times_out_at.isoformat(),
            )
            return None

    browser_address = browser_session.browser_address

    if not browser_address:
        LOG.info(
            "Waiting for browser session address.",
            browser_session_id=browser_session_id,
            organization_id=organization_id,
        )

        try:
            browser_address = await app.PERSISTENT_SESSIONS_MANAGER.get_browser_address(
                session_id=browser_session_id,
                organization_id=organization_id,
            )
        except Exception as ex:
            LOG.info(
                "Browser session address not found for browser session.",
                browser_session_id=browser_session_id,
                organization_id=organization_id,
                ex=ex,
            )
            return None

    try:
        addressable_browser_session = AddressablePersistentBrowserSession(
            **browser_session.model_dump() | {"browser_address": browser_address},
        )
    except Exception:
        return None

    return addressable_browser_session


async def verify_task(
    task_id: str, organization_id: str
) -> tuple[Task | None, AddressablePersistentBrowserSession | None]:
    """
    Verify the task is running, and that it has a browser session associated
    with it.
    """

    task = await app.DATABASE.get_task(task_id=task_id, organization_id=organization_id)

    if not task:
        LOG.info("Task not found.", task_id=task_id, organization_id=organization_id)
        return None, None

    if task.status.is_final():
        LOG.info("Task is in a final state.", task_status=task.status, task_id=task_id, organization_id=organization_id)

        return None, None

    if task.status not in [TaskStatus.created, TaskStatus.queued, TaskStatus.running]:
        LOG.info(
            "Task is not created, queued, or running.",
            task_status=task.status,
            task_id=task_id,
            organization_id=organization_id,
        )

        return None, None

    browser_session = await app.PERSISTENT_SESSIONS_MANAGER.get_session_by_runnable_id(
        organization_id=organization_id,
        runnable_id=task_id,
    )

    if not browser_session:
        LOG.info("No browser session found for task.", task_id=task_id, organization_id=organization_id)
        return task, None

    if not browser_session.browser_address:
        LOG.info("Browser session address not found for task.", task_id=task_id, organization_id=organization_id)
        return task, None

    try:
        addressable_browser_session = AddressablePersistentBrowserSession(
            **browser_session.model_dump() | {"browser_address": browser_session.browser_address},
        )
    except Exception as e:
        LOG.error("vnc.browser-session-reify-error", task_id=task_id, organization_id=organization_id, error=e)
        return task, None

    return task, addressable_browser_session


async def verify_workflow_run(
    workflow_run_id: str,
    organization_id: str,
) -> tuple[WorkflowRun | None, AddressablePersistentBrowserSession | None]:
    """
    Verify the workflow run is running, and that it has a browser session associated
    with it.
    """

    if settings.ENV == "local":
        dummy_workflow_run = WorkflowRun(
            workflow_id="123",
            workflow_permanent_id="wpid_123",
            workflow_run_id=workflow_run_id,
            organization_id=organization_id,
            status=WorkflowRunStatus.running,
            created_at=datetime.now(),
            modified_at=datetime.now(),
        )

        dummy_browser_session = AddressablePersistentBrowserSession(
            persistent_browser_session_id=workflow_run_id,
            organization_id=organization_id,
            browser_address="0.0.0.0:9223",
            ip_address="localhost",
            created_at=datetime.now(),
            modified_at=datetime.now(),
        )

        return dummy_workflow_run, dummy_browser_session

    workflow_run = await app.DATABASE.get_workflow_run(
        workflow_run_id=workflow_run_id,
        organization_id=organization_id,
    )

    if not workflow_run:
        LOG.info("Workflow run not found.", workflow_run_id=workflow_run_id, organization_id=organization_id)
        return None, None

    if workflow_run.status.is_final():
        LOG.info(
            "Workflow run is in a final state. Closing connection.",
            workflow_run_status=workflow_run.status,
            workflow_run_id=workflow_run_id,
            organization_id=organization_id,
        )

        return None, None

    if workflow_run.status not in [
        WorkflowRunStatus.created,
        WorkflowRunStatus.queued,
        WorkflowRunStatus.running,
        WorkflowRunStatus.paused,
    ]:
        LOG.info(
            "Workflow run is not running.",
            workflow_run_status=workflow_run.status,
            workflow_run_id=workflow_run_id,
            organization_id=organization_id,
        )

        return None, None

    browser_session = await app.PERSISTENT_SESSIONS_MANAGER.get_session_by_runnable_id(
        organization_id=organization_id,
        runnable_id=workflow_run_id,
    )

    if not browser_session:
        LOG.info(
            "No browser session found for workflow run.",
            workflow_run_id=workflow_run_id,
            organization_id=organization_id,
        )
        return workflow_run, None

    browser_address = browser_session.browser_address

    if not browser_address:
        LOG.info(
            "Waiting for browser session address.", workflow_run_id=workflow_run_id, organization_id=organization_id
        )

        try:
            browser_address = await app.PERSISTENT_SESSIONS_MANAGER.get_browser_address(
                session_id=browser_session.persistent_browser_session_id,
                organization_id=organization_id,
            )
        except Exception as ex:
            LOG.info(
                "Browser session address not found for workflow run.",
                workflow_run_id=workflow_run_id,
                organization_id=organization_id,
                ex=ex,
            )
            return workflow_run, None

    try:
        addressable_browser_session = AddressablePersistentBrowserSession(
            **browser_session.model_dump() | {"browser_address": browser_address},
        )
    except Exception:
        return workflow_run, None

    return workflow_run, addressable_browser_session


async def loop_verify_browser_session(verifiable: MessageChannel | VncChannel) -> None:
    """
    Loop until the browser session is cleared or the websocket is closed.
    """

    while verifiable.browser_session and verifiable.is_open:
        browser_session = await verify_browser_session(
            browser_session_id=verifiable.browser_session.persistent_browser_session_id,
            organization_id=verifiable.organization_id,
        )

        verifiable.browser_session = browser_session

        await asyncio.sleep(Constants.POLL_INTERVAL_FOR_VERIFICATION_SECONDS)


async def loop_verify_task(vnc_channel: VncChannel) -> None:
    """
    Loop until the task is cleared or the websocket is closed.
    """

    while vnc_channel.task and vnc_channel.is_open:
        task, browser_session = await verify_task(
            task_id=vnc_channel.task.task_id,
            organization_id=vnc_channel.organization_id,
        )

        vnc_channel.task = task
        vnc_channel.browser_session = browser_session

        await asyncio.sleep(Constants.POLL_INTERVAL_FOR_VERIFICATION_SECONDS)


async def loop_verify_workflow_run(verifiable: MessageChannel | VncChannel) -> None:
    """
    Loop until the workflow run is cleared or the websocket is closed.
    """

    while verifiable.workflow_run and verifiable.is_open:
        workflow_run, browser_session = await verify_workflow_run(
            workflow_run_id=verifiable.workflow_run.workflow_run_id,
            organization_id=verifiable.organization_id,
        )

        verifiable.workflow_run = workflow_run
        verifiable.browser_session = browser_session

        await asyncio.sleep(Constants.POLL_INTERVAL_FOR_VERIFICATION_SECONDS)



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .action import Action
from .block_type import BlockType
from .run_engine import RunEngine
from .workflow_run_block_data_schema import WorkflowRunBlockDataSchema
from .workflow_run_block_navigation_payload import WorkflowRunBlockNavigationPayload
from .workflow_run_block_output import WorkflowRunBlockOutput


class WorkflowRunBlock(UniversalBaseModel):
    workflow_run_block_id: str
    block_workflow_run_id: typing.Optional[str] = None
    workflow_run_id: str
    organization_id: str
    description: typing.Optional[str] = None
    parent_workflow_run_block_id: typing.Optional[str] = None
    block_type: BlockType
    label: typing.Optional[str] = None
    status: typing.Optional[str] = None
    output: typing.Optional[WorkflowRunBlockOutput] = None
    continue_on_failure: typing.Optional[bool] = None
    failure_reason: typing.Optional[str] = None
    engine: typing.Optional[RunEngine] = None
    task_id: typing.Optional[str] = None
    url: typing.Optional[str] = None
    navigation_goal: typing.Optional[str] = None
    navigation_payload: typing.Optional[WorkflowRunBlockNavigationPayload] = None
    data_extraction_goal: typing.Optional[str] = None
    data_schema: typing.Optional[WorkflowRunBlockDataSchema] = None
    terminate_criterion: typing.Optional[str] = None
    complete_criterion: typing.Optional[str] = None
    actions: typing.Optional[typing.List[Action]] = None
    created_at: dt.datetime
    modified_at: dt.datetime
    include_action_history_in_verification: typing.Optional[bool] = None
    duration: typing.Optional[float] = None
    loop_values: typing.Optional[typing.List[typing.Optional[typing.Any]]] = None
    current_value: typing.Optional[str] = None
    current_index: typing.Optional[int] = None
    recipients: typing.Optional[typing.List[str]] = None
    attachments: typing.Optional[typing.List[str]] = None
    subject: typing.Optional[str] = None
    body: typing.Optional[str] = None
    instructions: typing.Optional[str] = None
    positive_descriptor: typing.Optional[str] = None
    negative_descriptor: typing.Optional[str] = None
    executed_branch_id: typing.Optional[str] = None
    executed_branch_expression: typing.Optional[str] = None
    executed_branch_result: typing.Optional[bool] = None
    executed_branch_next_block: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""add debug_session_id to workflow_runs table

Revision ID: c50ee6f26432
Revises: ac9d96ea0501
Create Date: 2025-10-01 11:16:03.680233+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "c50ee6f26432"
down_revision: Union[str, None] = "ac9d96ea0501"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflow_runs", sa.Column("debug_session_id", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflow_runs", "debug_session_id")
    # ### end Alembic commands ###



from __future__ import annotations

import asyncio
import time
from enum import StrEnum
from io import BytesIO
from typing import Any

import structlog
from PIL import Image
from playwright._impl._errors import TimeoutError
from playwright.async_api import ElementHandle, Frame, Page

from skyvern.constants import PAGE_CONTENT_TIMEOUT, SKYVERN_DIR
from skyvern.exceptions import FailedToTakeScreenshot
from skyvern.forge.sdk.settings_manager import SettingsManager
from skyvern.forge.sdk.trace import traced

LOG = structlog.get_logger()


def load_js_script() -> str:
    # TODO: Handle file location better. This is a hacky way to find the file location.
    path = f"{SKYVERN_DIR}/webeye/scraper/domUtils.js"
    try:
        # TODO: Implement TS of domUtils.js and use the complied JS file instead of the raw JS file.
        # This will allow our code to be type safe.
        with open(path, encoding="utf-8") as f:
            return f.read()
    except FileNotFoundError as e:
        LOG.exception("Failed to load the JS script", path=path)
        raise e


JS_FUNCTION_DEFS = load_js_script()


class ScreenshotMode(StrEnum):
    LITE = "lite"
    DETAILED = "detailed"


async def _page_screenshot_helper(
    page: Page,
    file_path: str | None = None,
    full_page: bool = False,
    timeout: float = SettingsManager.get_settings().BROWSER_SCREENSHOT_TIMEOUT_MS,
) -> bytes:
    try:
        return await page.screenshot(
            path=file_path,
            timeout=timeout,
            full_page=full_page,
            animations="disabled",
        )
    except TimeoutError as timeout_error:
        LOG.info(
            f"Timeout error while taking screenshot: {str(timeout_error)}. Going to take a screenshot again with animation allowed."
        )
        return await page.screenshot(
            path=file_path,
            timeout=timeout,
            full_page=full_page,
            animations="allow",
        )


async def _current_viewpoint_screenshot_helper(
    page: Page,
    file_path: str | None = None,
    full_page: bool = False,
    timeout: float = SettingsManager.get_settings().BROWSER_SCREENSHOT_TIMEOUT_MS,
    mode: ScreenshotMode = ScreenshotMode.DETAILED,
) -> bytes:
    if page.is_closed():
        raise FailedToTakeScreenshot(error_message="Page is closed")

    # Capture page context for debugging screenshot issues
    url = page.url
    try:
        viewport = page.viewport_size
        viewport_info = f"{viewport['width']}x{viewport['height']}" if viewport else "unknown"
    except Exception:
        viewport_info = "unknown"

    try:
        if mode == ScreenshotMode.DETAILED:
            await page.wait_for_load_state(timeout=SettingsManager.get_settings().BROWSER_LOADING_TIMEOUT_MS)
            LOG.debug("Page is fully loaded, agent is about to take screenshots")
        start_time = time.time()
        screenshot: bytes = b""
        if file_path:
            screenshot = await _page_screenshot_helper(
                page=page, file_path=file_path, full_page=full_page, timeout=timeout
            )
        else:
            screenshot = await _page_screenshot_helper(page=page, full_page=full_page, timeout=timeout)
        end_time = time.time()
        LOG.debug(
            "Screenshot taking time",
            screenshot_time=end_time - start_time,
            file_path=file_path,
        )
        return screenshot
    except TimeoutError as e:
        LOG.error(
            "Screenshot timeout",
            timeout_ms=timeout,
            url=url,
            viewport=viewport_info,
            full_page=full_page,
            mode=mode.value if hasattr(mode, "value") else str(mode),
            error=str(e),
        )
        raise FailedToTakeScreenshot(error_message=str(e)) from e
    except Exception as e:
        LOG.error(
            "Screenshot failed",
            url=url,
            viewport=viewport_info,
            full_page=full_page,
            error=str(e),
            exc_info=True,
        )
        raise FailedToTakeScreenshot(error_message=str(e)) from e


async def _scrolling_screenshots_helper(
    page: Page,
    url: str | None = None,
    draw_boxes: bool = False,
    max_number: int = SettingsManager.get_settings().MAX_NUM_SCREENSHOTS,
    mode: ScreenshotMode = ScreenshotMode.DETAILED,
) -> tuple[list[bytes], list[int]]:
    # page is the main frame and the index must be 0
    skyvern_page = await SkyvernFrame.create_instance(frame=page)
    frame = "main.frame"
    frame_index = 0

    # when mode is lite, we don't draw bounding boxes
    # since draw_boxes impacts the performance of processing
    if mode == ScreenshotMode.LITE:
        draw_boxes = False

    screenshots: list[bytes] = []
    positions: list[int] = []
    if await skyvern_page.is_window_scrollable():
        scroll_y_px_old = -30.0
        _, initial_scroll_height = await skyvern_page.get_scroll_width_and_height()
        scroll_y_px = await skyvern_page.scroll_to_top(draw_boxes=draw_boxes, frame=frame, frame_index=frame_index)
        # Checking max number of screenshots to prevent infinite loop
        # We are checking the difference between the old and new scroll_y_px to determine if we have reached the end of the
        # page. If the difference is less than 25, we assume we have reached the end of the page.
        while abs(scroll_y_px_old - scroll_y_px) > 25 and len(screenshots) < max_number:
            # check if the scroll height changed, if so, rebuild the element tree
            _, scroll_height = await skyvern_page.get_scroll_width_and_height()
            if scroll_height != initial_scroll_height:
                LOG.warning(
                    "Scroll height changed, rebuild the element tree",
                    scroll_height=scroll_height,
                    initial_scroll_height=initial_scroll_height,
                )
                await skyvern_page.build_tree_from_body(frame_name=frame, frame_index=frame_index)
                initial_scroll_height = scroll_height

            screenshot = await _current_viewpoint_screenshot_helper(page=page, mode=mode)
            screenshots.append(screenshot)
            positions.append(int(scroll_y_px))
            scroll_y_px_old = scroll_y_px
            LOG.debug("Scrolling to next page", url=url, num_screenshots=len(screenshots))
            scroll_y_px = await skyvern_page.scroll_to_next_page(
                draw_boxes=draw_boxes,
                frame=frame,
                frame_index=frame_index,
                need_overlap=(mode == ScreenshotMode.DETAILED),
            )
            LOG.debug(
                "Scrolled to next page",
                scroll_y_px=scroll_y_px,
                scroll_y_px_old=scroll_y_px_old,
            )
        if draw_boxes:
            await skyvern_page.remove_bounding_boxes()
        await skyvern_page.scroll_to_top(draw_boxes=False, frame=frame, frame_index=frame_index)

        if mode == ScreenshotMode.DETAILED:
            # wait until animation ends, which is triggered by scrolling
            await skyvern_page.safe_wait_for_animation_end()
    else:
        if draw_boxes:
            await skyvern_page.build_elements_and_draw_bounding_boxes(frame=frame, frame_index=frame_index)

        LOG.debug("Page is not scrollable", url=url, num_screenshots=len(screenshots))
        screenshot = await _current_viewpoint_screenshot_helper(page=page, mode=mode)
        screenshots.append(screenshot)
        positions.append(0)

        if draw_boxes:
            await skyvern_page.remove_bounding_boxes()

    return screenshots, positions


def _merge_images_by_position(images: list[Image.Image], positions: list[int]) -> Image.Image:
    """Merge screenshots vertically using scroll positions to remove overlaps."""
    if not images:
        raise ValueError("no images to merge")
    if len(images) != len(positions):
        raise ValueError("images and positions length mismatch")

    if len(images) == 1:
        return images[0]

    max_width = max(img.width for img in images)

    merged_height = images[0].height
    for i in range(1, len(images)):
        merged_height += positions[i] - positions[i - 1]

    merged_img = Image.new("RGB", (max_width, merged_height), color=(255, 255, 255))

    current_y = 0
    merged_img.paste(images[0], (0, current_y))
    current_y += images[0].height

    for i in range(1, len(images)):
        step = positions[i] - positions[i - 1]
        overlap = images[i].height - step
        if overlap > 0:
            cropped = images[i].crop((0, overlap, images[i].width, images[i].height))
        else:
            cropped = images[i]

        merged_img.paste(cropped, (0, current_y))
        current_y += cropped.height

    return merged_img


class SkyvernFrame:
    @staticmethod
    async def evaluate(
        frame: Page | Frame,
        expression: str,
        arg: Any | None = None,
        timeout_ms: float = SettingsManager.get_settings().BROWSER_ACTION_TIMEOUT_MS,
    ) -> Any:
        try:
            async with asyncio.timeout(timeout_ms / 1000):
                return await frame.evaluate(expression=expression, arg=arg)
        except asyncio.TimeoutError:
            LOG.exception("Skyvern timed out trying to analyze the page", expression=expression)
            raise TimeoutError("Skyvern timed out trying to analyze the page")

    @staticmethod
    async def get_url(frame: Page | Frame) -> str:
        return await SkyvernFrame.evaluate(frame=frame, expression="() => document.location.href")

    @staticmethod
    @traced()
    async def take_scrolling_screenshot(
        page: Page,
        file_path: str | None = None,
        timeout: float = SettingsManager.get_settings().BROWSER_SCREENSHOT_TIMEOUT_MS,
        mode: ScreenshotMode = ScreenshotMode.DETAILED,
        scrolling_number: int = SettingsManager.get_settings().MAX_NUM_SCREENSHOTS,
    ) -> bytes:
        if scrolling_number <= 0:
            return await _current_viewpoint_screenshot_helper(
                page=page, file_path=file_path, timeout=timeout, mode=mode
            )

        if scrolling_number > SettingsManager.get_settings().MAX_NUM_SCREENSHOTS:
            LOG.warning(
                "scrolling_number is greater than the max number of screenshots, setting it to the max number of screenshots",
                scrolling_number=scrolling_number,
                max_number=SettingsManager.get_settings().MAX_NUM_SCREENSHOTS,
            )
            scrolling_number = SettingsManager.get_settings().MAX_NUM_SCREENSHOTS

        # use spilt screenshot with lite mode, isntead of fullpage screenshot from playwright
        LOG.debug("Page is fully loaded, agent is about to generate the full page screenshot")
        start_time = time.time()
        skyvern_frame = await SkyvernFrame.create_instance(frame=page)
        x: int | None = None
        y: int | None = None
        try:
            x, y = await skyvern_frame.get_scroll_x_y()
            async with asyncio.timeout(timeout):
                screenshots, positions = await _scrolling_screenshots_helper(
                    page=page, mode=mode, max_number=scrolling_number
                )
                images = []

                for screenshot in screenshots:
                    with Image.open(BytesIO(screenshot)) as img:
                        img.load()
                        images.append(img)

                merged_img = _merge_images_by_position(images, positions)

                buffer = BytesIO()
                merged_img.save(buffer, format="PNG")
                buffer.seek(0)

                img_data = buffer.read()
                if file_path is not None:
                    with open(file_path, "wb") as f:
                        f.write(img_data)

                end_time = time.time()
                LOG.debug(
                    "Full page screenshot taking time",
                    screenshot_time=end_time - start_time,
                    file_path=file_path,
                )
                return img_data
        except Exception:
            LOG.warning(
                "Failed to take full page screenshot, fallback to use playwright full page screenshot",
                exc_info=True,
            )
            # reset x and y to None to avoid the scroll_to_x_y call in finally block
            x = None
            y = None
            return await _current_viewpoint_screenshot_helper(
                page=page, file_path=file_path, timeout=timeout, full_page=True
            )
        finally:
            if x is not None and y is not None:
                await skyvern_frame.safe_scroll_to_x_y(x, y)

    @staticmethod
    @traced()
    async def take_split_screenshots(
        page: Page,
        url: str | None = None,
        draw_boxes: bool = False,
        max_number: int = SettingsManager.get_settings().MAX_NUM_SCREENSHOTS,
        scroll: bool = True,
    ) -> list[bytes]:
        if not scroll:
            return [await _current_viewpoint_screenshot_helper(page=page, mode=ScreenshotMode.DETAILED)]

        screenshots, _ = await _scrolling_screenshots_helper(
            page=page,
            url=url,
            max_number=max_number,
            draw_boxes=draw_boxes,
            mode=ScreenshotMode.DETAILED,
        )
        return screenshots

    @classmethod
    async def create_instance(cls, frame: Page | Frame) -> SkyvernFrame:
        instance = cls(frame=frame)
        await cls.evaluate(frame=instance.frame, expression=JS_FUNCTION_DEFS)
        if SettingsManager.get_settings().ENABLE_EXP_ALL_TEXTUAL_ELEMENTS_INTERACTABLE:
            await instance.evaluate(
                frame=instance.frame, expression="() => window.GlobalEnableAllTextualElements = true"
            )
        return instance

    def __init__(self, frame: Page | Frame) -> None:
        self.frame = frame

    def get_frame(self) -> Page | Frame:
        return self.frame

    async def get_content(self, timeout: float = PAGE_CONTENT_TIMEOUT) -> str:
        async with asyncio.timeout(timeout):
            return await self.frame.content()

    async def get_scroll_x_y(self) -> tuple[int, int]:
        js_script = "() => getScrollXY()"
        return await self.evaluate(frame=self.frame, expression=js_script)

    async def get_scroll_width_and_height(self) -> tuple[int, int]:
        js_script = "() => getScrollWidthAndHeight()"
        return await self.evaluate(frame=self.frame, expression=js_script)

    async def scroll_to_x_y(self, x: int, y: int) -> None:
        js_script = "([x, y]) => scrollToXY(x, y)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=[x, y])

    async def safe_scroll_to_x_y(self, x: int, y: int) -> None:
        try:
            await self.scroll_to_x_y(x, y)
        except Exception:
            LOG.warning("Failed to scroll to x, y, ignore it", x=x, y=y, exc_info=True)

    async def scroll_to_element_bottom(self, element: ElementHandle, page_by_page: bool = False) -> None:
        js_script = "([element, page_by_page]) => scrollToElementBottom(element, page_by_page)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=[element, page_by_page])

    async def scroll_to_element_top(self, element: ElementHandle) -> None:
        js_script = "(element) => scrollToElementTop(element)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=element)

    async def parse_element_from_html(self, frame: str, element: ElementHandle, interactable: bool) -> dict:
        js_script = "async ([frame, element, interactable]) => await buildElementObject(frame, element, interactable)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=[frame, element, interactable])

    async def get_element_scrollable(self, element: ElementHandle) -> bool:
        js_script = "(element) => isScrollable(element)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=element)

    async def get_element_visible(self, element: ElementHandle) -> bool:
        js_script = "(element) => isElementVisible(element) && !isHidden(element)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=element)

    async def get_disabled_from_style(self, element: ElementHandle) -> bool:
        js_script = "(element) => checkDisabledFromStyle(element)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=element)

    async def get_blocking_element_id(self, element: ElementHandle) -> tuple[str, bool]:
        js_script = "(element) => getBlockElementUniqueID(element)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=element)

    async def scroll_to_top(self, draw_boxes: bool, frame: str, frame_index: int) -> float:
        """
        Scroll to the top of the page and take a screenshot.
        :param drow_boxes: If True, draw bounding boxes around the elements.
        :param page: Page instance to take the screenshot from.
        :return: Screenshot of the page.
        """
        js_script = "async ([draw_boxes, frame, frame_index]) => await safeScrollToTop(draw_boxes, frame, frame_index)"
        scroll_y_px = await self.evaluate(
            frame=self.frame,
            expression=js_script,
            timeout_ms=SettingsManager.get_settings().BROWSER_SCRAPING_BUILDING_ELEMENT_TREE_TIMEOUT_MS,
            arg=[draw_boxes, frame, frame_index],
        )
        return scroll_y_px

    async def scroll_to_next_page(
        self, draw_boxes: bool, frame: str, frame_index: int, need_overlap: bool = True
    ) -> float:
        """
        Scroll to the next page and take a screenshot.
        :param drow_boxes: If True, draw bounding boxes around the elements.
        :param page: Page instance to take the screenshot from.
        :return: Screenshot of the page.
        """
        js_script = "async ([draw_boxes, frame, frame_index, need_overlap]) => await scrollToNextPage(draw_boxes, frame, frame_index, need_overlap)"
        scroll_y_px = await self.evaluate(
            frame=self.frame,
            expression=js_script,
            timeout_ms=SettingsManager.get_settings().BROWSER_SCRAPING_BUILDING_ELEMENT_TREE_TIMEOUT_MS,
            arg=[draw_boxes, frame, frame_index, need_overlap],
        )
        return scroll_y_px

    async def remove_bounding_boxes(self) -> None:
        """
        Remove the bounding boxes from the page.
        :param page: Page instance to remove the bounding boxes from.
        """
        js_script = "() => removeBoundingBoxes()"
        await self.evaluate(
            frame=self.frame,
            expression=js_script,
            timeout_ms=SettingsManager.get_settings().BROWSER_SCRAPING_BUILDING_ELEMENT_TREE_TIMEOUT_MS,
        )

    async def build_elements_and_draw_bounding_boxes(self, frame: str, frame_index: int) -> None:
        js_script = "async ([frame, frame_index]) => await buildElementsAndDrawBoundingBoxes(frame, frame_index)"
        await self.evaluate(
            frame=self.frame,
            expression=js_script,
            timeout_ms=SettingsManager.get_settings().BROWSER_SCRAPING_BUILDING_ELEMENT_TREE_TIMEOUT_MS,
            arg=[frame, frame_index],
        )

    async def is_window_scrollable(self) -> bool:
        js_script = "() => isWindowScrollable()"
        return await self.evaluate(frame=self.frame, expression=js_script)

    async def is_parent(self, parent: ElementHandle, child: ElementHandle) -> bool:
        js_script = "([parent, child]) => isParent(parent, child)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=[parent, child])

    async def is_sibling(self, el1: ElementHandle, el2: ElementHandle) -> bool:
        js_script = "([el1, el2]) => isSibling(el1, el2)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=[el1, el2])

    async def has_ASP_client_control(self) -> bool:
        js_script = "() => hasASPClientControl()"
        return await self.evaluate(frame=self.frame, expression=js_script)

    async def click_element_in_javascript(self, element: ElementHandle) -> None:
        js_script = "(element) => element.click()"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=element)

    async def remove_target_attr(self, element: ElementHandle) -> None:
        js_script = "(element) => element.removeAttribute('target')"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=element)

    async def get_select_options(self, element: ElementHandle) -> tuple[list, str]:
        js_script = "([element]) => getSelectOptions(element)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=[element])

    async def get_element_dom_depth(self, element: ElementHandle) -> int:
        js_script = "([element]) => getElementDomDepth(element)"
        return await self.evaluate(frame=self.frame, expression=js_script, arg=[element])

    async def remove_all_unique_ids(self) -> None:
        js_script = "() => removeAllUniqueIds()"
        await self.evaluate(frame=self.frame, expression=js_script)

    @traced()
    async def build_tree_from_body(
        self,
        frame_name: str | None,
        frame_index: int,
        must_included_tags: list[str] | None = None,
        timeout_ms: float = SettingsManager.get_settings().BROWSER_SCRAPING_BUILDING_ELEMENT_TREE_TIMEOUT_MS,
    ) -> tuple[list[dict], list[dict]]:
        must_included_tags = must_included_tags or []
        js_script = "async ([frame_name, frame_index, must_included_tags]) => await buildTreeFromBody(frame_name, frame_index, must_included_tags)"
        return await self.evaluate(
            frame=self.frame,
            expression=js_script,
            timeout_ms=timeout_ms,
            arg=[frame_name, frame_index, must_included_tags],
        )

    @traced()
    async def get_incremental_element_tree(
        self,
        wait_until_finished: bool = True,
        timeout_ms: float = SettingsManager.get_settings().BROWSER_SCRAPING_BUILDING_ELEMENT_TREE_TIMEOUT_MS,
    ) -> tuple[list[dict], list[dict]]:
        js_script = "async ([wait_until_finished]) => await getIncrementElements(wait_until_finished)"
        return await self.evaluate(
            frame=self.frame, expression=js_script, timeout_ms=timeout_ms, arg=[wait_until_finished]
        )

    @traced()
    async def build_tree_from_element(
        self,
        starter: ElementHandle,
        frame: str,
        full_tree: bool = False,
        timeout_ms: float = SettingsManager.get_settings().BROWSER_SCRAPING_BUILDING_ELEMENT_TREE_TIMEOUT_MS,
    ) -> tuple[list[dict], list[dict]]:
        js_script = "async ([starter, frame, full_tree]) => await buildElementTree(starter, frame, full_tree)"
        return await self.evaluate(
            frame=self.frame, expression=js_script, timeout_ms=timeout_ms, arg=[starter, frame, full_tree]
        )

    async def safe_wait_for_animation_end(self, before_wait_sec: float = 0, timeout_ms: float = 3000) -> None:
        try:
            await asyncio.sleep(before_wait_sec)
            await self.frame.wait_for_load_state("load", timeout=timeout_ms)
            await self.wait_for_animation_end(timeout_ms=timeout_ms)
        except Exception:
            LOG.debug("Failed to wait for animation end, but ignore it", exc_info=True)
            return

    async def wait_for_animation_end(self, timeout_ms: float = 3000) -> None:
        async with asyncio.timeout(timeout_ms / 1000):
            while True:
                is_finished = await self.evaluate(
                    frame=self.frame,
                    expression="() => isAnimationFinished()",
                    timeout_ms=timeout_ms,
                )
                if is_finished:
                    return
                await asyncio.sleep(0.1)

    async def wait_for_page_ready(
        self,
        network_idle_timeout_ms: float = 3000,
        loading_indicator_timeout_ms: float = 5000,
        dom_stable_ms: float = 300,
        dom_stability_timeout_ms: float = 3000,
    ) -> None:
        """
        Wait for page to be ready for interaction by checking multiple signals:
        1. Loading indicators gone (spinners, skeletons, progress bars) - highest timeout first
        2. Network idle (no pending requests for 500ms)
        3. DOM stability (no significant mutations for dom_stable_ms)

        Checks are ordered by timeout (highest first) so the longest timeout
        acts as the primary upper bound when checks complete early.

        This is designed for cached action execution to ensure the page is ready
        before attempting to interact with elements.
        """
        total_start_time = time.time()

        # 1. Wait for loading indicators to disappear (longest timeout first)
        loading_indicator_duration_ms = 0.0
        step_start_time = time.time()
        loading_indicator_result = "success"
        try:
            await self._wait_for_loading_indicators_gone(timeout_ms=loading_indicator_timeout_ms)
        except (TimeoutError, asyncio.TimeoutError):
            loading_indicator_result = "timeout"
            LOG.warning("Loading indicator timeout - some indicators may still be present, proceeding")
        except Exception:
            loading_indicator_result = "error"
            LOG.warning("Failed to check loading indicators, proceeding", exc_info=True)
        finally:
            loading_indicator_duration_ms = (time.time() - step_start_time) * 1000
            LOG.info(
                "page_readiness_check",
                step="loading_indicators",
                result=loading_indicator_result,
                duration_ms=loading_indicator_duration_ms,
                timeout_ms=loading_indicator_timeout_ms,
            )

        # 2. Wait for network idle (with short timeout - some pages never go idle)
        network_idle_duration_ms = 0.0
        step_start_time = time.time()
        network_idle_result = "success"
        try:
            await self.frame.wait_for_load_state("networkidle", timeout=network_idle_timeout_ms)
        except (TimeoutError, asyncio.TimeoutError):
            network_idle_result = "timeout"
            LOG.warning("Network idle timeout - page may have constant activity, proceeding")
        finally:
            network_idle_duration_ms = (time.time() - step_start_time) * 1000
            LOG.info(
                "page_readiness_check",
                step="network_idle",
                result=network_idle_result,
                duration_ms=network_idle_duration_ms,
                timeout_ms=network_idle_timeout_ms,
            )

        # 3. Wait for DOM to stabilize
        dom_stability_duration_ms = 0.0
        step_start_time = time.time()
        dom_stability_result = "success"
        try:
            await self._wait_for_dom_stable(stable_ms=dom_stable_ms, timeout_ms=dom_stability_timeout_ms)
        except (TimeoutError, asyncio.TimeoutError):
            dom_stability_result = "timeout"
            LOG.warning("DOM stability timeout - DOM may still be changing, proceeding")
        except Exception:
            dom_stability_result = "error"
            LOG.warning("Failed to check DOM stability, proceeding", exc_info=True)
        finally:
            dom_stability_duration_ms = (time.time() - step_start_time) * 1000
            LOG.info(
                "page_readiness_check",
                step="dom_stability",
                result=dom_stability_result,
                duration_ms=dom_stability_duration_ms,
                timeout_ms=dom_stability_timeout_ms,
                stable_ms=dom_stable_ms,
            )

        # Log total page readiness check duration
        total_duration_ms = (time.time() - total_start_time) * 1000
        LOG.info(
            "page_readiness_check_complete",
            total_duration_ms=total_duration_ms,
            loading_indicator_duration_ms=loading_indicator_duration_ms,
            network_idle_duration_ms=network_idle_duration_ms,
            dom_stability_duration_ms=dom_stability_duration_ms,
            loading_indicator_result=loading_indicator_result,
            network_idle_result=network_idle_result,
            dom_stability_result=dom_stability_result,
        )

    async def _wait_for_loading_indicators_gone(self, timeout_ms: float = 5000) -> None:
        """
        Wait for common loading indicators to disappear from the page.
        Checks for spinners, skeletons, progress bars, and loading overlays.
        """
        # JavaScript to detect loading indicators
        loading_indicator_js = """
        () => {
            // Common loading indicator selectors
            const selectors = [
                // Class-based spinners and loaders
                '[class*="spinner"]',
                '[class*="loading"]',
                '[class*="loader"]',
                '[class*="skeleton"]',
                '[class*="progress"]',
                '[class*="shimmer"]',
                // Role-based
                '[role="progressbar"]',
                '[role="status"][aria-busy="true"]',
                // Aria attributes
                '[aria-busy="true"]',
                '[aria-live="polite"][aria-busy="true"]',
                // Common loading overlay patterns
                '.loading-overlay',
                '.page-loading',
                '.content-loading',
                // SVG spinners
                'svg[class*="spin"]',
                'svg[class*="loading"]',
            ];

            for (const selector of selectors) {
                try {
                    const elements = document.querySelectorAll(selector);
                    for (const el of elements) {
                        // Check if element is visible
                        const style = window.getComputedStyle(el);
                        const rect = el.getBoundingClientRect();
                        const isVisible = (
                            style.display !== 'none' &&
                            style.visibility !== 'hidden' &&
                            style.opacity !== '0' &&
                            rect.width > 0 &&
                            rect.height > 0
                        );
                        if (isVisible) {
                            return true;  // Loading indicator found
                        }
                    }
                } catch (e) {
                    // Ignore selector errors
                }
            }
            return false;  // No loading indicators found
        }
        """

        async with asyncio.timeout(timeout_ms / 1000):
            while True:
                has_loading_indicator = await self.evaluate(
                    frame=self.frame,
                    expression=loading_indicator_js,
                    timeout_ms=timeout_ms,
                )
                if not has_loading_indicator:
                    LOG.debug("No loading indicators detected")
                    return
                await asyncio.sleep(0.1)

    async def _wait_for_dom_stable(self, stable_ms: float = 300, timeout_ms: float = 3000) -> None:
        """
        Wait for DOM to stabilize (no significant mutations for stable_ms milliseconds).
        Uses MutationObserver to detect DOM changes.
        """
        dom_stability_js = f"""
        () => new Promise((resolve) => {{
            let lastMutationTime = Date.now();
            let resolved = false;

            const observer = new MutationObserver((mutations) => {{
                // Filter out insignificant mutations (attribute changes on non-visible elements)
                const significantMutations = mutations.filter(m => {{
                    if (m.type === 'childList') return true;
                    if (m.type === 'characterData') return true;
                    if (m.type === 'attributes') {{
                        const el = m.target;
                        if (el.nodeType !== 1) return false;
                        const rect = el.getBoundingClientRect();
                        // Only count attribute changes on visible elements
                        return rect.width > 0 && rect.height > 0;
                    }}
                    return false;
                }});

                if (significantMutations.length > 0) {{
                    lastMutationTime = Date.now();
                }}
            }});

            observer.observe(document.body, {{
                childList: true,
                subtree: true,
                attributes: true,
                characterData: true,
            }});

            const checkStability = () => {{
                if (resolved) return;
                const timeSinceLastMutation = Date.now() - lastMutationTime;
                if (timeSinceLastMutation >= {stable_ms}) {{
                    resolved = true;
                    observer.disconnect();
                    resolve(true);
                }} else {{
                    setTimeout(checkStability, 50);
                }}
            }};

            // Start checking after a brief delay to catch initial mutations
            setTimeout(checkStability, 50);
        }})
        """

        await self.evaluate(
            frame=self.frame,
            expression=dom_stability_js,
            timeout_ms=timeout_ms,
        )



import structlog
from fastapi import HTTPException

from skyvern.forge import app
from skyvern.forge.sdk.schemas.credentials import (
    CreateCredentialRequest,
    Credential,
    CredentialItem,
    CredentialVaultType,
)
from skyvern.forge.sdk.services.bitwarden import BitwardenService
from skyvern.forge.sdk.services.credential.credential_vault_service import CredentialVaultService

LOG = structlog.get_logger()


class BitwardenCredentialVaultService(CredentialVaultService):
    async def create_credential(self, organization_id: str, data: CreateCredentialRequest) -> Credential:
        org_collection = await app.DATABASE.get_organization_bitwarden_collection(organization_id)

        if not org_collection:
            LOG.info(
                "There is no collection for the organization. Creating new collection.",
                organization_id=organization_id,
            )
            collection_id = await BitwardenService.create_collection(
                name=organization_id,
            )
            org_collection = await app.DATABASE.create_organization_bitwarden_collection(
                organization_id,
                collection_id,
            )

        item_id = await BitwardenService.create_credential_item(
            collection_id=org_collection.collection_id,
            name=data.name,
            credential=data.credential,
        )

        credential = await self._create_db_credential(
            organization_id=organization_id,
            data=data,
            item_id=item_id,
            vault_type=CredentialVaultType.BITWARDEN,
        )

        return credential

    async def update_credential(self, credential: Credential, data: CreateCredentialRequest) -> Credential:
        org_collection = await app.DATABASE.get_organization_bitwarden_collection(credential.organization_id)

        if not org_collection:
            raise HTTPException(status_code=404, detail="Credential account not found. It might have been deleted.")

        # Create new vault item with the updated data
        new_item_id = await BitwardenService.create_credential_item(
            collection_id=org_collection.collection_id,
            name=data.name,
            credential=data.credential,
        )

        # Update DB record to point to the new vault item
        try:
            updated_credential = await self._update_db_credential(
                credential=credential,
                data=data,
                item_id=new_item_id,
            )
        except Exception:
            LOG.warning(
                "DB update failed, attempting to clean up new Bitwarden vault item",
                organization_id=credential.organization_id,
                new_item_id=new_item_id,
            )
            try:
                await BitwardenService.delete_credential_item(new_item_id)
            except Exception as cleanup_error:
                LOG.error(
                    "Failed to clean up orphaned Bitwarden vault item",
                    organization_id=credential.organization_id,
                    new_item_id=new_item_id,
                    error=str(cleanup_error),
                )
            raise

        return updated_credential

    async def delete_credential(
        self,
        credential: Credential,
    ) -> None:
        organization_bitwarden_collection = await app.DATABASE.get_organization_bitwarden_collection(
            credential.organization_id
        )
        if not organization_bitwarden_collection:
            raise HTTPException(status_code=404, detail="Credential account not found. It might have been deleted.")

        await app.DATABASE.delete_credential(credential.credential_id, credential.organization_id)
        await BitwardenService.delete_credential_item(credential.item_id)

    async def post_delete_credential_item(self, item_id: str, organization_id: str | None = None) -> None:
        try:
            await BitwardenService.delete_credential_item(item_id)
            LOG.info(
                "Successfully deleted credential item from Bitwarden in background",
                item_id=item_id,
            )
        except Exception as e:
            LOG.warning(
                "Failed to delete credential item from Bitwarden in background",
                item_id=item_id,
                error=str(e),
                exc_info=True,
            )

    async def get_credential_item(self, db_credential: Credential) -> CredentialItem:
        return await BitwardenService.get_credential_item(db_credential.item_id)



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .branch_condition_yaml import BranchConditionYaml


class ConditionalBlockYaml(UniversalBaseModel):
    label: str = pydantic.Field()
    """
    Author-facing identifier; must be unique per workflow.
    """

    next_block_label: typing.Optional[str] = pydantic.Field(default=None)
    """
    Optional pointer to the label of the next block. When omitted, it will default to sequential order. See [[s-4bnl]].
    """

    continue_on_failure: typing.Optional[bool] = None
    model: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None
    next_loop_on_failure: typing.Optional[bool] = None
    branch_conditions: typing.Optional[typing.List[BranchConditionYaml]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
